from okta.cache.cache import Cache
from okta.cache.no_op_cache import NoOpCache
from okta.cache.okta_cache import OktaCache
from tests.mocks import mock_pause_function, mock_cache_return_none,\
    mock_cache_return_value, TTI, TTL, CACHE_KEY, CACHE_VALUE,\
    ALT_CACHE_KEY, ALT_CACHE_VALUE

import pytest
import time


@pytest.fixture(scope="module")
def setup(monkeypatch):
    monkeypatch.setattr(time, 'sleep', mock_pause_function)


def test_cache_key_creation():
    cache = Cache()
    new_key = cache.create_key(CACHE_KEY)
    assert new_key is not None
    assert new_key == CACHE_KEY


def test_cache_add_entry():
    cache = OktaCache(TTL, TTI)
    cache.add(CACHE_KEY, CACHE_VALUE)
    assert cache._store[CACHE_KEY]["value"] == CACHE_VALUE

    cache.add("test string", CACHE_VALUE)
    assert cache._store["test string"]["value"] == CACHE_VALUE


def test_cache_has_key():
    cache = OktaCache(TTL, TTI)

    assert(not cache.contains(CACHE_KEY))

    cache.add(CACHE_KEY, CACHE_VALUE)

    assert(cache.contains(CACHE_KEY))
    assert(not cache.contains(ALT_CACHE_KEY))


def test_cache_get_value():
    cache = OktaCache(TTL, TTI)

    assert(cache.get(CACHE_KEY) is None)
    cache.add(CACHE_KEY, CACHE_VALUE)
    assert(cache.get(CACHE_KEY) is CACHE_VALUE)


def test_cache_delete_value():
    cache = OktaCache(TTL, TTI)
    cache.add(CACHE_KEY, CACHE_VALUE)

    assert(cache.contains(CACHE_KEY))
    cache.delete(CACHE_KEY)
    assert(not cache.contains(CACHE_KEY))


def test_cache_clear():
    cache = OktaCache(TTL, TTI)

    first_key = CACHE_KEY
    first_value = CACHE_VALUE

    second_key = ALT_CACHE_KEY
    second_value = ALT_CACHE_VALUE

    cache.add(first_key, first_value)
    cache.add(second_key, second_value)
    assert(cache.contains(first_key) and cache.contains(second_key))
    cache.clear()
    assert(not (cache.contains(first_key) or cache.contains(second_key)))


def test_cache_TTI(monkeypatch):
    local_TTL = float(10)
    local_TTI = float(1)
    cache = OktaCache(local_TTL, local_TTI)

    assert(cache.get(CACHE_KEY) is None)
    cache.add(CACHE_KEY, CACHE_VALUE)
    time.sleep(local_TTI / 2)
    monkeypatch.setattr(OktaCache, 'get', mock_cache_return_value)
    assert(cache.get(CACHE_KEY) is CACHE_VALUE)  # resets TTI
    time.sleep(local_TTI)
    monkeypatch.setattr(OktaCache, 'get', mock_cache_return_none)
    assert(cache.get(CACHE_KEY) is None)


def test_cache_TTL(monkeypatch):
    local_TTL = float(2)
    local_TTI = float(10)
    cache = OktaCache(local_TTL, local_TTI)
    KEY = "12345"
    VALUE = "54321"

    assert(cache.get(KEY) is None)
    cache.add(KEY, VALUE)
    time.sleep(local_TTL)
    monkeypatch.setattr(OktaCache, 'get', mock_cache_return_none)
    assert(cache.get(KEY) is None)  # deleted by now

    cache.add(KEY, VALUE)
    time.sleep(local_TTL / 2)
    monkeypatch.setattr(OktaCache, 'get', mock_cache_return_value)
    assert(cache.get(KEY) is VALUE)
    time.sleep(local_TTL / 2)
    monkeypatch.setattr(OktaCache, 'get', mock_cache_return_none)
    assert(cache.get(KEY) is None)  # deleted by now


@pytest.mark.parametrize("value", [None, "string", 1, 1.0, True])
def test_no_op_cache_for_nothingness(value):
    cache = NoOpCache()

    assert cache.contains(value) is False
    cache.add(value, value)
    assert cache.get(value) is None



from urllib.parse import urlparse


class Cache():
    """
    This is the ABSTRACT class that defines a Cache object for the Okta Client
    """

    def __init__(self):
        pass

    def get(self, key):
        """
        A method which retrieves the desired value from the cache.

        Arguments:
            key {str} -- The key used to find the desired value

        Raises:
            NotImplementedError: If the subclass inheriting this class
            has not implemented this function
        """
        raise NotImplementedError

    def contains(self, key):
        """
        A method which checks if the cache contains the desired value.

        Arguments:
            key {str} -- The key used to check the desired value

        Raises:
            NotImplementedError: If the subclass inheriting this class
            has not implemented this function
        """
        raise NotImplementedError

    def add(self, key, value):
        """
        A method which adds a key-value pair to the cache.

        Arguments:
            key {str} -- The key used to identify the entry.
            value {[type]} -- The value in the pair

        Raises:
            NotImplementedError: If the subclass inheriting this class
            has not implemented this function
        """
        raise NotImplementedError

    def delete(self, key):
        """
        A method which deletes a key-value pair from the cache.

        Arguments:
            key {str} -- The key used to identify the entry

        Raises:
            NotImplementedError: If the subclass inheriting this class
            has not implemented this function
        """
        raise NotImplementedError

    def clear(self):
        """
        A method used to empty the cache.

        Raises:
            NotImplementedError: If the subclass inheriting this class
            has not implemented this function
        """
        raise NotImplementedError

    def create_key(self, request):
        """
        A method used to create a unique key for an entry in the cache.
        Used with URLs that requests fire at.

        Arguments:
            request {str} -- The key to use to produce a unique key

        Returns:
            str -- Unique key based on the input
        """
        # Validate URL and return URL string
        url_object = urlparse(request)
        return url_object.geturl()



import pytest
from tests.mocks import MockOktaClient
import okta.models as models
from http import HTTPStatus
from okta.errors.okta_api_error import OktaAPIError


class TestInlineHooksResource:
    """
    Integration Tests for the Inline Hooks Resource
    """
    SDK_PREFIX = "python_sdk"

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_create_get_inline_hook(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Inline Hook
        inline_hook_model = models.InlineHook({
            "name": f"{TestInlineHooksResource.SDK_PREFIX} Test Inline Hook",
            "version": "1.0.0",
            "type": models.InlineHookType.COM_OKTA_OAUTH_2_TOKENS_TRANSFORM,
            "channel": models.InlineHookChannel({
                "type": "HTTP",
                "version": "1.0.0",
                "config": models.InlineHookChannelConfig({
                    "uri": "https://www.example.com/inlineHook",
                    "headers": [
                        models.InlineHookChannelConfigHeaders({
                            "key": "X-Test-Header",
                            "value": "Test Header Value"
                        })
                    ],
                    "authScheme": models.InlineHookChannelConfigAuthScheme({
                        "type": "HEADER",
                        "key": "Authorization",
                        "value": "Test-Api-Key"
                    })
                })
            })
        })

        created_inline_hook, _, err = await \
            client.create_inline_hook(inline_hook_model)
        assert err is None
        assert isinstance(created_inline_hook, models.InlineHook)
        assert created_inline_hook.id
        assert created_inline_hook.name == inline_hook_model.name
        assert created_inline_hook.channel
        assert created_inline_hook.channel.config
        assert created_inline_hook.channel.config.uri ==\
            inline_hook_model.channel.config.uri

        # Retrieve
        retrieved_inline_hook, _, err = await \
            client.get_inline_hook(created_inline_hook.id)
        assert err is None
        assert isinstance(retrieved_inline_hook, models.InlineHook)
        assert retrieved_inline_hook.id == created_inline_hook.id
        assert retrieved_inline_hook.name == created_inline_hook.name

        # Deactivate & Delete
        deactivated_inline_hook, _, err = await \
            client.deactivate_inline_hook(created_inline_hook.id)
        assert err is None
        assert isinstance(deactivated_inline_hook, models.InlineHook)
        assert deactivated_inline_hook.status == "INACTIVE"

        _, err = await client.delete_inline_hook(created_inline_hook.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_update_inline_hook(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Inline Hook
        inline_hook_model = models.InlineHook({
            "name": f"{TestInlineHooksResource.SDK_PREFIX} Test Inline Hook",
            "version": "1.0.0",
            "type": models.InlineHookType.COM_OKTA_OAUTH_2_TOKENS_TRANSFORM,
            "channel": models.InlineHookChannel({
                "type": "HTTP",
                "version": "1.0.0",
                "config": models.InlineHookChannelConfig({
                    "uri": "https://www.example.com/inlineHook",
                    "headers": [
                        models.InlineHookChannelConfigHeaders({
                            "key": "X-Test-Header",
                            "value": "Test Header Value"
                        })
                    ],
                    "authScheme": models.InlineHookChannelConfigAuthScheme({
                        "type": "HEADER",
                        "key": "Authorization",
                        "value": "Test-Api-Key"
                    })
                })
            })
        })

        created_inline_hook, _, err = await \
            client.create_inline_hook(inline_hook_model)
        assert err is None
        assert isinstance(created_inline_hook, models.InlineHook)
        assert created_inline_hook.id
        assert created_inline_hook.name == inline_hook_model.name
        assert created_inline_hook.channel
        assert created_inline_hook.channel.config
        assert created_inline_hook.channel.config.uri ==\
            inline_hook_model.channel.config.uri

        # Update
        updated_inline_hook_model = models.InlineHook({
            "name": inline_hook_model.name + "UPDATE",
            "version": "1.0.0",
            "type": models.InlineHookType.COM_OKTA_OAUTH_2_TOKENS_TRANSFORM,
            "channel": models.InlineHookChannel({
                "type": "HTTP",
                "version": "1.0.0",
                "config": models.InlineHookChannelConfig({
                    "uri": inline_hook_model.channel.config.uri + "UPDATE",
                    "headers": [
                        models.InlineHookChannelConfigHeaders({
                            "key": "X-Test-Header",
                            "value": "Test Header Value UPDATE"
                        })
                    ],
                    "authScheme": models.InlineHookChannelConfigAuthScheme({
                        "type": "HEADER",
                        "key": "Authorization",
                        "value": "Test-Api-Key-UPDATE"
                    })
                })
            })
        })
        updated_inline_hook, _, err = await \
            client.update_inline_hook(
                created_inline_hook.id, updated_inline_hook_model)
        assert err is None
        assert updated_inline_hook.id == created_inline_hook.id
        assert updated_inline_hook.name == updated_inline_hook_model.name
        assert updated_inline_hook.channel.config.uri == \
            updated_inline_hook_model.channel.config.uri

        # Retrieve
        retrieved_inline_hook, _, err = await \
            client.get_inline_hook(created_inline_hook.id)
        assert err is None
        assert isinstance(retrieved_inline_hook, models.InlineHook)
        assert retrieved_inline_hook.id == created_inline_hook.id
        assert retrieved_inline_hook.name == updated_inline_hook.name
        assert retrieved_inline_hook.version == created_inline_hook.version
        assert retrieved_inline_hook.type ==\
            models.InlineHookType.COM_OKTA_OAUTH_2_TOKENS_TRANSFORM
        assert retrieved_inline_hook.channel.config.uri ==\
            updated_inline_hook.channel.config.uri

        # Deactivate & Delete
        deactivated_inline_hook, _, err = await \
            client.deactivate_inline_hook(created_inline_hook.id)
        assert err is None
        assert isinstance(deactivated_inline_hook, models.InlineHook)
        assert deactivated_inline_hook.status == "INACTIVE"

        _, err = await client.delete_inline_hook(created_inline_hook.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_delete_inline_hook(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Inline Hook
        inline_hook_model = models.InlineHook({
            "name": f"{TestInlineHooksResource.SDK_PREFIX} Test Inline Hook",
            "version": "1.0.0",
            "type": models.InlineHookType.COM_OKTA_OAUTH_2_TOKENS_TRANSFORM,
            "channel": models.InlineHookChannel({
                "type": "HTTP",
                "version": "1.0.0",
                "config": models.InlineHookChannelConfig({
                    "uri": "https://www.example.com/inlineHook",
                    "headers": [
                        models.InlineHookChannelConfigHeaders({
                            "key": "X-Test-Header",
                            "value": "Test Header Value"
                        })
                    ],
                    "authScheme": models.InlineHookChannelConfigAuthScheme({
                        "type": "HEADER",
                        "key": "Authorization",
                        "value": "Test-Api-Key"
                    })
                })
            })
        })

        created_inline_hook, _, err = await \
            client.create_inline_hook(inline_hook_model)
        assert err is None
        assert isinstance(created_inline_hook, models.InlineHook)
        assert created_inline_hook.id
        assert created_inline_hook.name == inline_hook_model.name
        assert created_inline_hook.channel
        assert created_inline_hook.channel.config
        assert created_inline_hook.channel.config.uri ==\
            inline_hook_model.channel.config.uri

        # Deactivate & Delete
        deactivated_inline_hook, _, err = await \
            client.deactivate_inline_hook(created_inline_hook.id)
        assert err is None
        assert isinstance(deactivated_inline_hook, models.InlineHook)
        assert deactivated_inline_hook.status == "INACTIVE"

        _, err = await client.delete_inline_hook(created_inline_hook.id)
        assert err is None

        # Retrieve
        retrieved_inline_hook, resp, err = await \
            client.get_inline_hook(created_inline_hook.id)
        assert err is not None
        assert isinstance(err, OktaAPIError)
        assert resp.get_status() == HTTPStatus.NOT_FOUND
        assert retrieved_inline_hook is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_activate_deactivate_inline_hook(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Inline Hook
        inline_hook_model = models.InlineHook({
            "name": f"{TestInlineHooksResource.SDK_PREFIX} Test Inline Hook",
            "version": "1.0.0",
            "type": models.InlineHookType.COM_OKTA_OAUTH_2_TOKENS_TRANSFORM,
            "channel": models.InlineHookChannel({
                "type": "HTTP",
                "version": "1.0.0",
                "config": models.InlineHookChannelConfig({
                    "uri": "https://www.example.com/inlineHook",
                    "headers": [
                        models.InlineHookChannelConfigHeaders({
                            "key": "X-Test-Header",
                            "value": "Test Header Value"
                        })
                    ],
                    "authScheme": models.InlineHookChannelConfigAuthScheme({
                        "type": "HEADER",
                        "key": "Authorization",
                        "value": "Test-Api-Key"
                    })
                })
            })
        })

        created_inline_hook, _, err = await \
            client.create_inline_hook(inline_hook_model)
        assert err is None
        assert isinstance(created_inline_hook, models.InlineHook)
        assert created_inline_hook.id
        assert created_inline_hook.name == inline_hook_model.name
        assert created_inline_hook.channel
        assert created_inline_hook.channel.config
        assert created_inline_hook.channel.config.uri ==\
            inline_hook_model.channel.config.uri
        assert created_inline_hook.status == "ACTIVE"

        # Deactivate
        deactivated_inline_hook, _, err = await \
            client.deactivate_inline_hook(created_inline_hook.id)
        assert err is None
        assert isinstance(deactivated_inline_hook, models.InlineHook)
        assert deactivated_inline_hook.status == "INACTIVE"

        # Retrieve
        retrieved_inline_hook, _, err = await \
            client.get_inline_hook(created_inline_hook.id)
        assert err is None
        assert retrieved_inline_hook.status == "INACTIVE"

        # Activate
        activated_inline_hook, _, err = await \
            client.activate_inline_hook(created_inline_hook.id)
        assert err is None
        assert isinstance(activated_inline_hook, models.InlineHook)
        assert activated_inline_hook.status == "ACTIVE"

        # Retrieve
        retrieved_inline_hook, _, err = await \
            client.get_inline_hook(created_inline_hook.id)
        assert err is None
        assert retrieved_inline_hook.status == "ACTIVE"

        # Delete & Deactivate
        _, _, err = await \
            client.deactivate_inline_hook(created_inline_hook.id)
        assert err is None
        _, err = await client.delete_inline_hook(created_inline_hook.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_list_inline_hooks(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Inline Hooks
        NUMBER_OF_HOOKS = 3
        created_inline_hooks = []
        for index in range(NUMBER_OF_HOOKS):
            inline_hook_model = models.InlineHook({
                "name": f"{TestInlineHooksResource.SDK_PREFIX} IH{index}",
                "version": "1.0.0",
                "type": models.InlineHookType.
                COM_OKTA_OAUTH_2_TOKENS_TRANSFORM,
                "channel": models.InlineHookChannel({
                    "type": "HTTP",
                    "version": "1.0.0",
                    "config": models.InlineHookChannelConfig({
                        "uri": "https://www.example.com/inlineHook",
                        "headers": [
                            models.InlineHookChannelConfigHeaders({
                                "key": "X-Test-Header",
                                "value": f"Test Header {index}"
                            })
                        ],
                        "authScheme": models.
                        InlineHookChannelConfigAuthScheme({
                            "type": "HEADER",
                            "key": "Authorization",
                            "value": f"Test-Api-Key-{index}"
                        })
                    })
                })
            })

            created_inline_hook, _, err = await \
                client.create_inline_hook(inline_hook_model)
            assert err is None
            assert isinstance(created_inline_hook, models.InlineHook)
            assert created_inline_hook.id
            assert created_inline_hook.name == inline_hook_model.name
            assert created_inline_hook.channel
            assert created_inline_hook.channel.config
            assert created_inline_hook.channel.config.uri ==\
                inline_hook_model.channel.config.uri
            created_inline_hooks.append(created_inline_hook.id)

        # List
        all_inline_hooks, _, err = await client.list_inline_hooks()
        assert err is None
        assert isinstance(all_inline_hooks, list)
        for inline_hook_id in created_inline_hooks:
            assert next((ih for ih in all_inline_hooks
                         if ih.id == inline_hook_id))

        # Deactivate & Delete
        for inline_hook_id in created_inline_hooks:
            _, _, err = await \
                client.deactivate_inline_hook(inline_hook_id)
            assert err is None
            _, err = await client.delete_inline_hook(inline_hook_id)
            assert err is None



import pytest
from tests.mocks import MockOktaClient
from tests.mocks import mock_pause_function
from http import HTTPStatus
import okta.models as models


class TestGroupsResource:
    """
    Integration Tests for the Groups Resource
    """

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_create_get_group(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Group Object
        GROUP_NAME = "Group-Target-Test"
        group_profile = models.GroupProfile({
            "name": GROUP_NAME
        })
        group_obj = models.Group({
            "profile": group_profile
        })

        # Create Group
        group, _, err = await client.create_group(group_obj)
        assert err is None
        assert isinstance(group, models.Group)

        # Get group using ID
        found_group, _, err = await client.get_group(group.id)
        assert err is None
        assert found_group.id == group.id

        # Delete created group
        _, err = await client.delete_group(group.id)
        assert err is None

        # Ensure group cannot be found again
        found_group, resp, err = await client.get_group(group.id)
        assert err is not None
        assert resp.get_status() == HTTPStatus.NOT_FOUND
        assert found_group is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_list_groups(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Group Object
        GROUP_NAME = "Group-Target-Test"
        group_profile = models.GroupProfile({
            "name": GROUP_NAME
        })
        group_obj = models.Group({
            "profile": group_profile
        })

        # Create Group
        group, _, err = await client.create_group(group_obj)
        assert err is None
        assert isinstance(group, models.Group)

        groups_list, resp, err = await client.list_groups()
        assert err is None
        assert not resp.has_next()
        assert next((grp for grp in groups_list if grp.id == group.id))

        # Delete created group
        _, err = await client.delete_group(group.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_search_group(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Group Object
        GROUP_NAME = "Group-Target-Test"
        group_profile = models.GroupProfile({
            "name": GROUP_NAME
        })
        group_obj = models.Group({
            "profile": group_profile
        })

        # Create Group
        group, _, err = await client.create_group(group_obj)
        assert err is None
        assert isinstance(group, models.Group)

        query_params_query = {"q": GROUP_NAME}
        groups_list, _, err = await client.list_groups(query_params_query)
        assert err is None
        assert groups_list
        assert len(groups_list) == 1
        assert next((grp for grp in groups_list if grp.id == group.id))

        # Delete created group
        _, err = await client.delete_group(group.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_update_group(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Group Object
        GROUP_NAME = "Group-Target-Test"
        group_profile = models.GroupProfile({
            "name": GROUP_NAME
        })
        group_obj = models.Group({
            "profile": group_profile
        })

        # Create Group
        group, _, err = await client.create_group(group_obj)
        assert err is None
        assert isinstance(group, models.Group)

        # Create Updated Group Object
        # Create Group Object
        NEW_GROUP_NAME = "Group-Target-Test NEW"
        new_group_profile = models.GroupProfile({
            "name": NEW_GROUP_NAME
        })
        new_group_obj = models.Group({
            "profile": new_group_profile
        })

        _, _, err = await client.update_group(group.id, new_group_obj)
        assert err is None

        # Verify update worked
        found_group, _, err = await client.get_group(group.id)
        assert err is None
        assert found_group.id == group.id
        assert found_group.profile.name == NEW_GROUP_NAME

        # Delete created group
        _, err = await client.delete_group(group.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_remove_group(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Password
        password = models.PasswordCredential({
            "value": "Password150kta"
        })
        # Create User Credentials
        user_creds = models.UserCredentials({
            "password": password
        })

        # Create User Profile and CreateUser Request
        user_profile = models.UserProfile()
        user_profile.first_name = "John"
        user_profile.last_name = "Doe-Remove-Group"
        user_profile.email = "John.Doe-Remove-Group@example.com"
        user_profile.login = "John.Doe-Remove-Group@example.com"

        create_user_req = models.CreateUserRequest({
            "credentials": user_creds,
            "profile": user_profile
        })

        # Create query parameters and Create User
        query_params_create = {"activate": "False"}
        user, _, err = await client.create_user(
            create_user_req, query_params_create)
        assert err is None
        assert isinstance(user, models.User)

        # Create Group Object
        GROUP_NAME = "Group-Target-Test"
        group_profile = models.GroupProfile({
            "name": GROUP_NAME
        })
        group_obj = models.Group({
            "profile": group_profile
        })

        # Create Group
        group, _, err = await client.create_group(group_obj)
        assert err is None
        assert isinstance(group, models.Group)

        # Add user to group
        _, err = await client.add_user_to_group(group.id, user.id)
        assert err is None

        users_in_group, _, err = await client.list_group_users(group.id)
        assert err is None
        assert next((usr for usr in users_in_group if usr.id == user.id))

        # Delete created group
        _, err = await client.delete_group(group.id)
        assert err is None

        # Retrieve group list again to ensure deleted
        groups_list, _, err = await client.list_groups()
        assert err is None
        assert next((grp for grp in groups_list if grp.id ==
                     group.id), None) is None

        # Deactivate, then delete created user
        _, err = await client.deactivate_or_delete_user(user.id)
        assert err is None
        _, err = await client.deactivate_or_delete_user(user.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_group_roles_operations(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Group Object
        GROUP_NAME = "Group-Target-Test"
        group_profile = models.GroupProfile({
            "name": GROUP_NAME
        })
        group_obj = models.Group({
            "profile": group_profile
        })

        # Create Group
        group, _, err = await client.create_group(group_obj)
        assert err is None
        assert isinstance(group, models.Group)

        # Create roles
        assign_role_req_ua = models.AssignRoleRequest({
            "type": models.RoleType.USER_ADMIN
        })
        assign_role_req_aa = models.AssignRoleRequest({
            "type": models.RoleType.APP_ADMIN
        })

        ua_role, _, err = await client.assign_role_to_group(
            group.id, assign_role_req_ua)
        assert err is None
        aa_role, _, err = await client.assign_role_to_group(
            group.id, assign_role_req_aa)
        assert err is None

        group_roles, _, err = await client.list_group_assigned_roles(group.id)
        assert err is None
        assert len(group_roles) == 2
        assert next((rle for rle in group_roles if rle.id == ua_role.id))
        assert next((rle for rle in group_roles if rle.id == aa_role.id))

        _, err = await client.remove_role_from_group(group.id, ua_role.id)
        assert err is None

        group_roles, _, err = await client.list_group_assigned_roles(group.id)
        assert err is None
        assert len(group_roles) == 1
        assert next((rle for rle in group_roles if rle.id ==
                     ua_role.id), None) is None
        assert next((rle for rle in group_roles if rle.id == aa_role.id))

        # Delete created group
        _, err = await client.delete_group(group.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_group_users_operations(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Password
        password = models.PasswordCredential({
            "value": "Password150kta"
        })
        # Create User Credentials
        user_creds = models.UserCredentials({
            "password": password
        })

        # Create User Profile and CreateUser Request
        user_profile = models.UserProfile()
        user_profile.first_name = "John"
        user_profile.last_name = "Doe-Activate"
        user_profile.email = "John.Doe-Activate@example.com"
        user_profile.login = "John.Doe-Activate@example.com"

        create_user_req = models.CreateUserRequest({
            "credentials": user_creds,
            "profile": user_profile
        })

        # Create query parameters and Create User
        query_params_create = {"activate": "False"}
        user, _, err = await client.create_user(
            create_user_req, query_params_create)
        assert err is None
        assert isinstance(user, models.User)

        # Create Group Object
        GROUP_NAME = "Group-Target-Test"
        group_profile = models.GroupProfile({
            "name": GROUP_NAME
        })
        group_obj = models.Group({
            "profile": group_profile
        })

        # Create Group
        group, _, err = await client.create_group(group_obj)
        assert err is None
        assert isinstance(group, models.Group)

        # Add user to group
        _, err = await client.add_user_to_group(group.id, user.id)
        assert err is None

        users_in_group, _, err = await client.list_group_users(group.id)
        assert err is None
        assert next((usr for usr in users_in_group if usr.id == user.id))

        # Remove user from group
        _, err = await client.remove_user_from_group(group.id, user.id)
        assert err is None

        users_in_group, _, err = await client.list_group_users(group.id)
        assert err is None
        assert len(users_in_group) == 0
        assert next(
            (usr for usr in users_in_group if usr.id == user.id), None) is None

        # Deactivate, then delete created user
        _, err = await client.deactivate_or_delete_user(user.id)
        assert err is None
        _, err = await client.deactivate_or_delete_user(user.id)
        assert err is None

        # Delete created group
        _, err = await client.delete_group(group.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_group_rule_operations(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Password
        password = models.PasswordCredential({
            "value": "Password150kta"
        })
        # Create User Credentials
        user_creds = models.UserCredentials({
            "password": password
        })

        # Create User Profile and CreateUser Request
        user_profile = models.UserProfile()
        user_profile.first_name = "John"
        user_profile.last_name = "Doe-Group-Rule-Ops"
        user_profile.email = "John.Doe-Group-Rule-Ops@example.com"
        user_profile.login = "John.Doe-Group-Rule-Ops@example.com"

        create_user_req = models.CreateUserRequest({
            "credentials": user_creds,
            "profile": user_profile
        })

        # Create query parameters and Create User
        query_params_create = {"activate": "False"}
        user, _, err = await client.create_user(
            create_user_req, query_params_create)
        assert err is None
        assert isinstance(user, models.User)

        # Create Group Object
        GROUP_NAME = "Group-Target-Test-Group-Rule-Ops"
        group_profile = models.GroupProfile({
            "name": GROUP_NAME
        })
        group_obj = models.Group({
            "profile": group_profile
        })

        # Create Group
        group, _, err = await client.create_group(group_obj)
        assert err is None
        assert isinstance(group, models.Group)

        # Create Group Rule
        last_name = user.profile.last_name
        GROUP_RULE_NAME = "Test-Group-Rule-Group-Rule-Ops"
        GROUP_RULE_TYPE = "group_rule"
        GROUP_RULE_EXP_TYPE = "urn:okta:expression:1.0"
        GROUP_RULE_EXP_VALUE = f"user.lastName==\"{last_name}\""
        group_rule_exp = models.GroupRuleExpression({
            "type": GROUP_RULE_EXP_TYPE,
            "value": GROUP_RULE_EXP_VALUE
        })

        group_rule_cond = models.GroupRuleConditions({
            "expression": group_rule_exp
        })

        group_rule_group_assignment = models.GroupRuleGroupAssignment({
            "groupIds": [group.id]
        })

        group_rule_action = models.GroupRuleAction({
            "assignUserToGroups": group_rule_group_assignment
        })

        group_rule_object = models.GroupRule({
            "actions": group_rule_action,
            "conditions": group_rule_cond,
            "name": GROUP_RULE_NAME,
            "type": GROUP_RULE_TYPE
        })

        group_rule, _, err = await client.create_group_rule(group_rule_object)
        assert err is None
        assert isinstance(group_rule, models.GroupRule)

        # Activate Group Rule
        _, err = await client.activate_group_rule(group_rule.id)
        assert err is None

        # 15 second sleep for backend to update
        mock_pause_function(15)

        users_in_group, _, err = await client.list_group_users(group.id)
        assert err is None
        assert next((usr for usr in users_in_group if usr.id ==
                     user.id), None) is not None

        # Ensure activated rule is in group rules
        group_rules, _, err = await client.list_group_rules()
        assert err is None
        assert next((rule for rule in group_rules if rule.id ==
                     group_rule.id), None) is not None

        # Deactivate rule (to update)
        _, err = await client.deactivate_group_rule(group_rule.id)
        assert err is None

        # Update rule
        # Create new rule
        NEW_GROUP_RULE_NAME = "Test-Group-Rule Updated"
        NEW_GROUP_RULE_EXP_VALUE = "user.lastName==\"BLAHBLAHBLAH\""
        new_group_rule_exp = models.GroupRuleExpression({
            "type": GROUP_RULE_EXP_TYPE,
            "value": NEW_GROUP_RULE_EXP_VALUE
        })

        new_group_rule_cond = models.GroupRuleConditions({
            "expression": new_group_rule_exp
        })

        new_group_rule_group_assignment = models.GroupRuleGroupAssignment({
            "groupIds": [group.id]
        })

        new_group_rule_action = models.GroupRuleAction({
            "assignUserToGroups": new_group_rule_group_assignment
        })

        new_group_rule_object = models.GroupRule({
            "actions": new_group_rule_action,
            "conditions": new_group_rule_cond,
            "name": NEW_GROUP_RULE_NAME,
            "type": GROUP_RULE_TYPE
        })

        new_group_rule, _, err = await client.update_group_rule(
            group_rule.id,
            new_group_rule_object)
        assert err is None

        # Activate updated rule and verify user isn't in group
        _, err = await client.activate_group_rule(new_group_rule.id)
        assert err is None

        # 15 second sleep for backend to update
        mock_pause_function(15)

        users_in_group, _, err = await client.list_group_users(group.id)
        assert err is None
        assert next(
            (usr for usr in users_in_group if usr.id == user.id), None) is None

        # Deactivate rule
        _, err = await client.deactivate_group_rule(new_group_rule.id)
        assert err is None

        # Deactivate, then delete created user
        _, err = await client.deactivate_user(user.id)
        assert err is None
        _, err = await client.deactivate_or_delete_user(user.id)
        assert err is None

        # Delete created group
        _, err = await client.delete_group(group.id)
        assert err is None

        # Delete group rules
        _, err = await client.delete_group_rule(group_rule.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_group_target_add(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Group Objects
        GROUP_1_NAME = "Group-Target-Test 1"
        group_1_profile = models.GroupProfile({
            "name": GROUP_1_NAME
        })
        group_1_obj = models.Group({
            "profile": group_1_profile
        })

        GROUP_2_NAME = "Group-Target-Test 2"
        group_2_profile = models.GroupProfile({
            "name": GROUP_2_NAME
        })
        group_2_obj = models.Group({
            "profile": group_2_profile
        })

        # Create Groups
        group_1, _, err = await client.create_group(group_1_obj)
        assert err is None
        assert isinstance(group_1, models.Group)

        group_2, _, err = await client.create_group(group_2_obj)
        assert err is None
        assert isinstance(group_2, models.Group)

        # Create role and add group targets
        assign_role_req_ua = models.AssignRoleRequest({
            "type": models.RoleType.USER_ADMIN
        })

        ua_role, _, err = await client.assign_role_to_group(
            group_1.id, assign_role_req_ua)
        assert err is None

        _, err = await\
            client.add_group_target_to_group_administrator_role_for_group(
                group_1.id, ua_role.id, group_2.id)

        # Make sure targets are listed
        groups_list, _, err = await client.list_group_targets_for_group_role(
            group_1.id, ua_role.id)
        assert err is None
        assert next((grp for grp in groups_list if grp.id == group_2.id))

        # Delete created groups
        _, err = await client.delete_group(group_1.id)
        assert err is None
        _, err = await client.delete_group(group_2.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_group_target_remove(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Group Objects
        GROUP_1_NAME = "Group-Target-Test 1"
        group_1_profile = models.GroupProfile({
            "name": GROUP_1_NAME
        })
        group_1_obj = models.Group({
            "profile": group_1_profile
        })

        GROUP_2_NAME = "Group-Target-Test 2"
        group_2_profile = models.GroupProfile({
            "name": GROUP_2_NAME
        })
        group_2_obj = models.Group({
            "profile": group_2_profile
        })

        GROUP_3_NAME = "Group-Target-Test 3"
        group_3_profile = models.GroupProfile({
            "name": GROUP_3_NAME
        })
        group_3_obj = models.Group({
            "profile": group_3_profile
        })

        # Create Groups
        group_1, _, err = await client.create_group(group_1_obj)
        assert err is None
        assert isinstance(group_1, models.Group)

        group_2, _, err = await client.create_group(group_2_obj)
        assert err is None
        assert isinstance(group_2, models.Group)

        group_3, _, err = await client.create_group(group_3_obj)
        assert err is None
        assert isinstance(group_3, models.Group)

        # Create role and add group targets
        assign_role_req_ua = models.AssignRoleRequest({
            "type": models.RoleType.USER_ADMIN
        })

        ua_role, _, err = await client.assign_role_to_group(
            group_1.id, assign_role_req_ua)
        assert err is None

        _, err = await\
            client.add_group_target_to_group_administrator_role_for_group(
                group_1.id, ua_role.id, group_2.id)
        _, err = await\
            client.add_group_target_to_group_administrator_role_for_group(
                group_1.id, ua_role.id, group_3.id)

        groups_list, _, err = await client.list_group_targets_for_group_role(
            group_1.id, ua_role.id)
        assert err is None
        assert next((grp for grp in groups_list if grp.id == group_2.id))
        assert next((grp for grp in groups_list if grp.id == group_3.id))

        # Remove from 2 and ensure 2 isn't listed
        _, err = await \
            client.remove_group_target_from_group_admin_role_given_to_group(
                group_1.id, ua_role.id, group_2.id)

        groups_list, _, err = await client.list_group_targets_for_group_role(
            group_1.id, ua_role.id)
        assert err is None
        assert next((grp for grp in groups_list if grp.id ==
                     group_2.id), None) is None
        assert next((grp for grp in groups_list if grp.id == group_3.id))

        # Delete created groups
        _, err = await client.delete_group(group_1.id)
        assert err is None
        _, err = await client.delete_group(group_2.id)
        assert err is None
        _, err = await client.delete_group(group_3.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_group_assigned_applications(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Group Objects
        GROUP_NAME = "Group-Target-Test"
        group_profile = models.GroupProfile({
            "name": GROUP_NAME
        })
        group_obj = models.Group({
            "profile": group_profile
        })

        # Create Group
        group, _, err = await client.create_group(group_obj)
        assert err is None
        assert isinstance(group, models.Group)

        # Create Application object and Application in Org
        APP_LABEL = "Test Assigned-Applications"
        BUTTON_FIELD = "btn-login"
        PASSWORD_FIELD = "txt-box-password"
        USERNAME_FIELD = "txt-box-username"
        URL = "https://example.com/login.html"
        LOGIN_URL_REGEX = f"^{URL}$"

        swa_app_settings_app = models.SwaApplicationSettingsApplication({
            "buttonField": BUTTON_FIELD,
            "passwordField": PASSWORD_FIELD,
            "usernameField": USERNAME_FIELD,
            "url": URL,
            "loginUrlRegex": LOGIN_URL_REGEX
        })

        swa_app_settings = models.SwaApplicationSettings({
            "app": swa_app_settings_app
        })

        swa_app_obj = models.SwaApplication({
            "label": APP_LABEL,
            "settings": swa_app_settings,
            "signOnMode": models.ApplicationSignOnMode.BROWSER_PLUGIN
        })

        swa_app, _, err = await client.create_application(swa_app_obj)
        assert err is None
        assert isinstance(swa_app, models.SwaApplication)

        # Assign app and group
        assign_ag_req = models.ApplicationGroupAssignment({
            "priority": 0,
            "applicationId": swa_app.id,
            "groupId": group.id
        })

        assign_app_group, _, err = await \
            client.create_application_group_assignment(
                swa_app.id, group.id, assign_ag_req)
        assert err is None

        # 3 second sleep for backend to update
        mock_pause_function(3)

        # Check assigned apps and ensure created app is found
        assigned_apps, _, err = await \
            client.list_assigned_applications_for_group(group.id)
        assert err is None
        assert assigned_apps is not None
        assert len(assigned_apps) > 0
        assert next((app for app in assigned_apps if app.id == swa_app.id))

        # Cleanup app and group created
        _, err = await client.deactivate_application(swa_app.id)
        assert err is None
        _, err = await client.delete_application(swa_app.id)
        assert err is None
        _, err = await client.delete_group(group.id)
        assert err is None



"""
Class of utility functions.
"""

from enum import Enum
from datetime import datetime as dt
from okta.constants import DATETIME_FORMAT, EPOCH_DAY, EPOCH_MONTH,\
    EPOCH_YEAR


def format_url(base_string):
    """
    Turns multiline strings in generated clients into
    simple one-line string

    Args:
        base_string (str): multiline URL

    Returns:
        str: single line URL
    """
    return ''.join(
        [line.strip() for line in base_string.splitlines()]
    )


def is_enum_value(value: str, enum: Enum):
    """
    Determines if given string is in given enumeration

    Args:
        value (str): string to test
        enum (Enum): enumeration to test on

    Returns:
        bool: True iff the string is in the enumeration
    """
    return value in set(entry.value for entry in enum)


def convert_date_time_to_seconds(date_time):
    """
    Takes in a date time string and returns the number of seconds
    since the epoch (Unix timestamp).

    Args:
        date_time (str): Date time string in the datetime format

    Returns:
        float: Number of seconds since the epoch
    """
    dt_obj = dt.strptime(date_time,
                         DATETIME_FORMAT)
    return float((dt_obj
                  - dt(EPOCH_YEAR, EPOCH_MONTH, EPOCH_DAY))
                 .total_seconds())



import pytest
from tests.mocks import MockOktaClient
import okta.models as models
from http import HTTPStatus
from okta.errors.okta_api_error import OktaAPIError


class TestPoliciesResource:
    """
    Integration Tests for the Policies Resource
    """
    SDK_PREFIX = "python_sdk"

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_create_get_sign_on_policy(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Policy
        policy_model = models.OktaSignOnPolicy({
            "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Sign-On",
            "status": "ACTIVE",
            "description": "Test policy applies for tests only"
        })

        created_policy, _, err = await client.create_policy(policy_model)
        assert err is None
        assert isinstance(created_policy, models.OktaSignOnPolicy)
        assert created_policy.name == policy_model.name
        assert created_policy.description == policy_model.description
        assert created_policy.status == "ACTIVE"
        assert created_policy.type == models.PolicyType.OKTA_SIGN_ON

        # Retrieve
        retrieved_policy, _, err = await client.get_policy(created_policy.id)
        assert err is None
        assert retrieved_policy.id == created_policy.id
        assert retrieved_policy.name == created_policy.name
        assert retrieved_policy.description == created_policy.description
        assert retrieved_policy.status == created_policy.status
        assert retrieved_policy.type == created_policy.type

        # Delete
        _, err = await client.delete_policy(created_policy.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_create_password_policy(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Policy
        policy_model = models.PasswordPolicy({
            "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Password",
            "status": "ACTIVE",
            "description": "Test policy applies for tests only"
        })

        created_policy, _, err = await client.create_policy(policy_model)
        assert err is None
        assert isinstance(created_policy, models.PasswordPolicy)
        assert created_policy.name == policy_model.name
        assert created_policy.description == policy_model.description
        assert created_policy.status == "ACTIVE"
        assert created_policy.type == models.PolicyType.PASSWORD

        # Retrieve
        retrieved_policy, _, err = await client.get_policy(created_policy.id)
        assert err is None
        assert retrieved_policy.id == created_policy.id
        assert retrieved_policy.name == created_policy.name
        assert retrieved_policy.description == created_policy.description
        assert retrieved_policy.status == created_policy.status
        assert retrieved_policy.type == created_policy.type

        # Delete
        _, err = await client.delete_policy(created_policy.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_create_get_policy(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Policy
        policy_model = models.Policy({
            "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Sign-On",
            "type": models.PolicyType.OKTA_SIGN_ON,
            "status": "ACTIVE",
            "description": "Test policy applies for tests only"
        })

        created_policy, _, err = await client.create_policy(policy_model)
        assert err is None
        assert isinstance(created_policy, models.OktaSignOnPolicy)
        assert created_policy.name == policy_model.name
        assert created_policy.description == policy_model.description
        assert created_policy.status == "ACTIVE"
        assert created_policy.type == models.PolicyType.OKTA_SIGN_ON

        # Retrieve
        retrieved_policy, _, err = await client.get_policy(created_policy.id)
        assert err is None
        assert retrieved_policy.id == created_policy.id
        assert retrieved_policy.name == created_policy.name
        assert retrieved_policy.description == created_policy.description
        assert retrieved_policy.status == created_policy.status
        assert retrieved_policy.type == created_policy.type

        # Delete
        _, err = await client.delete_policy(created_policy.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_create_get_sign_on_policy_with_group_conditions(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Group
        GROUP_NAME = "Group-Target-Test"
        group_obj = models.Group({
            "profile": models.GroupProfile({
                "name": GROUP_NAME
            })
        })
        created_group, _, err = await client.create_group(group_obj)
        assert err is None
        assert isinstance(created_group, models.Group)

        # Create Policy & Conditions
        policy_model = models.OktaSignOnPolicy({
            "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Sign-On",
            "status": "ACTIVE",
            "description": "Test policy applies for tests only"
        })

        policy_conditions = models.OktaSignOnPolicyConditions({
            "people": models.PolicyPeopleCondition({
                "groups": models.GroupCondition({
                    "include": [created_group.id]
                })
            })
        })

        policy_model.conditions = policy_conditions

        created_policy, _, err = await client.create_policy(policy_model)
        assert err is None
        assert isinstance(created_policy, models.OktaSignOnPolicy)
        assert created_policy.name == policy_model.name
        assert created_policy.description == policy_model.description
        assert created_policy.status == "ACTIVE"
        assert created_policy.type == models.PolicyType.OKTA_SIGN_ON
        assert len(created_policy.conditions.people.groups.include) == 1
        assert created_group.id in \
            created_policy.conditions.people.groups.include

        # Retrieve
        retrieved_policy, _, err = await client.get_policy(created_policy.id)
        assert err is None
        assert retrieved_policy.id == created_policy.id
        assert retrieved_policy.name == created_policy.name
        assert retrieved_policy.description == created_policy.description
        assert retrieved_policy.status == created_policy.status
        assert retrieved_policy.type == created_policy.type
        assert len(retrieved_policy.conditions.people.groups.include) == 1
        assert created_group.id in \
            retrieved_policy.conditions.people.groups.include

        # Delete Policy + Group
        _, err = await client.delete_policy(created_policy.id)
        assert err is None
        _, err = await client.delete_group(created_group.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_list_policies_by_type(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Policies
        policy_model = models.OktaSignOnPolicy({
            "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Sign-On",
            "status": "ACTIVE",
            "description": "Test sign on policy applies for tests only"
        })

        policy_model_2 = models.PasswordPolicy({
            "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Password",
            "status": "ACTIVE",
            "description": "Test password policy applies for tests only"
        })

        created_oss_policy, _, err = await client.create_policy(policy_model)
        assert err is None
        assert isinstance(created_oss_policy, models.OktaSignOnPolicy)
        assert created_oss_policy.name == policy_model.name
        assert created_oss_policy.description == policy_model.description
        assert created_oss_policy.status == "ACTIVE"
        assert created_oss_policy.type == models.PolicyType.OKTA_SIGN_ON

        created_pw_policy, _, err = await client.create_policy(policy_model_2)
        assert err is None
        assert isinstance(created_pw_policy, models.PasswordPolicy)
        assert created_pw_policy.name == policy_model_2.name
        assert created_pw_policy.description == policy_model_2.description
        assert created_pw_policy.status == "ACTIVE"
        assert created_pw_policy.type == models.PolicyType.PASSWORD

        # List by type
        query_params_list_sign_on = {"type": "OKTA_SIGN_ON"}
        sign_on_policies, _, err = await \
            client.list_policies(query_params_list_sign_on)
        assert err is None
        assert isinstance(sign_on_policies, list)
        assert next((plcy for plcy in sign_on_policies
                     if plcy.id == created_oss_policy.id))

        query_params_list_pw = {"type": "PASSWORD"}
        sign_on_policies, _, err = await \
            client.list_policies(query_params_list_pw)
        assert err is None
        assert isinstance(sign_on_policies, list)
        assert next((plcy for plcy in sign_on_policies
                     if plcy.id == created_pw_policy.id))

        # Delete
        _, err = await client.delete_policy(created_oss_policy.id)
        assert err is None
        _, err = await client.delete_policy(created_pw_policy.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_delete_policy(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Policy
        policy_model = models.OktaSignOnPolicy({
            "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Sign-On",
            "status": "ACTIVE",
            "description": "Test policy applies for tests only"
        })

        created_policy, _, err = await client.create_policy(policy_model)
        assert err is None
        assert isinstance(created_policy, models.OktaSignOnPolicy)
        assert created_policy.name == policy_model.name
        assert created_policy.description == policy_model.description
        assert created_policy.status == "ACTIVE"
        assert created_policy.type == models.PolicyType.OKTA_SIGN_ON

        # Retrieve
        retrieved_policy, _, err = await client.get_policy(created_policy.id)
        assert err is None
        assert retrieved_policy.id == created_policy.id
        assert retrieved_policy.name == created_policy.name
        assert retrieved_policy.description == created_policy.description
        assert retrieved_policy.status == created_policy.status
        assert retrieved_policy.type == created_policy.type

        # Delete
        _, err = await client.delete_policy(created_policy.id)
        assert err is None

        # Retrieve
        retrieved_policy, resp, err = await \
            client.get_policy(created_policy.id)
        assert err is not None
        assert isinstance(err, OktaAPIError)
        assert resp.get_status() == HTTPStatus.NOT_FOUND
        assert retrieved_policy is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_update_policy(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Policy
        policy_model = models.OktaSignOnPolicy({
            "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Sign-On",
            "status": "ACTIVE",
            "description": "Test policy applies for tests only"
        })

        created_policy, _, err = await client.create_policy(policy_model)
        assert err is None
        assert isinstance(created_policy, models.OktaSignOnPolicy)
        assert created_policy.name == policy_model.name
        assert created_policy.description == policy_model.description
        assert created_policy.status == "ACTIVE"
        assert created_policy.type == models.PolicyType.OKTA_SIGN_ON

        # Retrieve
        retrieved_policy, _, err = await client.get_policy(created_policy.id)
        assert err is None
        assert retrieved_policy.id == created_policy.id
        assert retrieved_policy.name == created_policy.name
        assert retrieved_policy.description == created_policy.description
        assert retrieved_policy.status == created_policy.status
        assert retrieved_policy.type == created_policy.type

        # Update
        NEW_NAME = policy_model.name + "UPDATED"
        NEW_DESC = policy_model.description + "UPDATED"
        created_policy.name = NEW_NAME
        created_policy.description = NEW_DESC
        updated_policy, _, err = await client.update_policy(
            created_policy.id, created_policy
        )
        assert err is None
        assert updated_policy.id == created_policy.id
        assert updated_policy.name == NEW_NAME
        assert updated_policy.description == NEW_DESC

        # Retrieve
        retrieved_policy, _, err = await client.get_policy(created_policy.id)
        assert err is None
        assert retrieved_policy.id == created_policy.id
        assert retrieved_policy.name == NEW_NAME
        assert retrieved_policy.description == NEW_DESC
        assert retrieved_policy.status == created_policy.status
        assert retrieved_policy.type == created_policy.type

        # Delete
        _, err = await client.delete_policy(created_policy.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_activate_deactivate_policy(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Policy
        policy_model = models.OktaSignOnPolicy({
            "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Sign-On",
            "status": "ACTIVE",
            "description": "Test policy applies for tests only"
        })

        created_policy, _, err = await client.create_policy(policy_model)
        assert err is None
        assert isinstance(created_policy, models.OktaSignOnPolicy)
        assert created_policy.name == policy_model.name
        assert created_policy.description == policy_model.description
        assert created_policy.status == "ACTIVE"
        assert created_policy.type == models.PolicyType.OKTA_SIGN_ON

        # Deactivate
        _, err = await client.deactivate_policy(created_policy.id)
        assert err is None

        # Retrieve
        retrieved_policy, _, err = await client.get_policy(created_policy.id)
        assert err is None
        assert retrieved_policy.id == created_policy.id
        assert retrieved_policy.name == created_policy.name
        assert retrieved_policy.description == created_policy.description
        assert retrieved_policy.status == "INACTIVE"
        assert retrieved_policy.type == created_policy.type

        # Reactivate
        _, err = await client.activate_policy(created_policy.id)
        assert err is None

        # Delete
        _, err = await client.delete_policy(created_policy.id)
        assert err is None

    """
    POLICY RULES TESTS BELOW
    """
    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_create_get_sign_on_policy_rule(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Policy
        policy_model = models.OktaSignOnPolicy({
            "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Sign-On",
            "status": "ACTIVE",
            "description": "Test policy applies for tests only"
        })

        created_policy, _, err = await client.create_policy(policy_model)
        assert err is None
        assert isinstance(created_policy, models.OktaSignOnPolicy)
        assert created_policy.name == policy_model.name
        assert created_policy.description == policy_model.description
        assert created_policy.status == "ACTIVE"
        assert created_policy.type == models.PolicyType.OKTA_SIGN_ON

        # Create Policy Rule
        policy_rule_model = models.OktaSignOnPolicyRule({
            "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Sign-On-Rule",
            "actions": models.OktaSignOnPolicyRuleActions({
                "signon": models.OktaSignOnPolicyRuleSignonActions({
                    "access": "ALLOW",
                    "requireFactor": "false",
                    "rememberDeviceByDefault": "false",
                    "session":
                    models.OktaSignOnPolicyRuleSignonSessionActions({
                        "usePersistentCookie": "false",
                        "maxSessionIdleMinutes": 720,
                        "maxSessionLifetimeMinutes": 0
                    })
                })
            }),
            "conditions": models.OktaSignOnPolicyRuleConditions({
                "authContext": models.PolicyRuleAuthContextCondition({
                    "authType": "ANY"
                })
            })
        })
        created_policy_rule, _, err = await client.create_policy_rule(
            created_policy.id, policy_rule_model
        )
        assert err is None
        assert isinstance(created_policy_rule, models.OktaSignOnPolicyRule)
        assert created_policy_rule.name == policy_rule_model.name
        assert created_policy_rule.type == "SIGN_ON"
        assert created_policy_rule.actions.signon.access == "ALLOW"
        assert created_policy_rule.actions.signon.require_factor is False
        assert created_policy_rule.actions.signon.remember_device_by_default\
            is False
        assert created_policy_rule.actions.signon.session.\
            use_persistent_cookie is False
        assert created_policy_rule.actions.signon.session.\
            max_session_idle_minutes == 720
        assert created_policy_rule.actions.signon.session.\
            max_session_lifetime_minutes == 0
        assert created_policy_rule.conditions.auth_context.auth_type == "ANY"

        # Retrieve
        retrieved_policy_rule, _, err = await client.get_policy_rule(
            created_policy.id, created_policy_rule.id)
        assert err is None
        assert retrieved_policy_rule.id == created_policy_rule.id
        assert retrieved_policy_rule.name == created_policy_rule.name
        assert isinstance(retrieved_policy_rule, models.OktaSignOnPolicyRule)

        # Delete
        _, err = await \
            client.delete_policy_rule(
                created_policy.id, created_policy_rule.id)
        assert err is None
        _, err = await client.delete_policy(created_policy.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_create_get_password_policy_rule(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Policy
        policy_model = models.PasswordPolicy({
            "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Password",
            "status": "ACTIVE",
            "description": "Test policy applies for tests only"
        })

        created_policy, _, err = await client.create_policy(policy_model)
        assert err is None
        assert isinstance(created_policy, models.PasswordPolicy)
        assert created_policy.name == policy_model.name
        assert created_policy.description == policy_model.description
        assert created_policy.status == "ACTIVE"
        assert created_policy.type == models.PolicyType.PASSWORD

        # Create Policy Rule
        policy_rule_model = models.PasswordPolicyRule({
            "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Password-Rule",
            "actions": models.PasswordPolicyRuleActions({
                "passwordChange": models.PasswordPolicyRuleAction({
                    "access": "ALLOW"
                }),
                "selfServicePasswordReset": models.PasswordPolicyRuleAction({
                    "access": "ALLOW"
                }),
                "selfServiceUnlock": models.PasswordPolicyRuleAction({
                    "access": "ALLOW"
                })
            }),
            "conditions": models.PasswordPolicyRuleConditions({
                "people": models.PolicyPeopleCondition({
                    "users": models.UserCondition({
                        "exclude": []
                    })
                }),
                "network": models.PolicyNetworkCondition({
                    "connection": "ANYWHERE"
                })
            })
        })
        created_policy_rule, _, err = await client.create_policy_rule(
            created_policy.id, policy_rule_model
        )
        assert err is None
        assert isinstance(created_policy_rule, models.PasswordPolicyRule)
        assert created_policy_rule.name == policy_rule_model.name
        assert created_policy_rule.type == "PASSWORD"
        assert created_policy_rule.status == "ACTIVE"
        assert created_policy_rule.conditions.people.users.exclude == []
        assert created_policy_rule.conditions.network.connection ==\
            "ANYWHERE"
        assert created_policy_rule.actions.password_change.access == "ALLOW"
        assert created_policy_rule.actions.self_service_password_reset\
            .access == "ALLOW"
        assert created_policy_rule.actions.self_service_unlock\
            .access == "ALLOW"

        # Retrieve
        retrieved_policy_rule, _, err = await client.get_policy_rule(
            created_policy.id, created_policy_rule.id)
        assert err is None
        assert retrieved_policy_rule.id == created_policy_rule.id
        assert retrieved_policy_rule.name == created_policy_rule.name
        assert isinstance(retrieved_policy_rule, models.PasswordPolicyRule)

        # Delete
        _, err = await \
            client.delete_policy_rule(
                created_policy.id, created_policy_rule.id)
        assert err is None
        _, err = await client.delete_policy(created_policy.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_update_policy_rule(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Policy
        policy_model = models.OktaSignOnPolicy({
            "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Sign-On",
            "status": "ACTIVE",
            "description": "Test policy applies for tests only"
        })

        created_policy, _, err = await client.create_policy(policy_model)
        assert err is None
        assert isinstance(created_policy, models.OktaSignOnPolicy)
        assert created_policy.name == policy_model.name
        assert created_policy.description == policy_model.description
        assert created_policy.status == "ACTIVE"
        assert created_policy.type == models.PolicyType.OKTA_SIGN_ON

        # Create Policy Rule
        policy_rule_model = models.OktaSignOnPolicyRule({
            "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Sign-On-Rule",
            "actions": models.OktaSignOnPolicyRuleActions({
                "signon": models.OktaSignOnPolicyRuleSignonActions({
                    "access": "ALLOW",
                    "requireFactor": "false",
                    "rememberDeviceByDefault": "false",
                    "session":
                    models.OktaSignOnPolicyRuleSignonSessionActions({
                        "usePersistentCookie": "false",
                        "maxSessionIdleMinutes": 720,
                        "maxSessionLifetimeMinutes": 0
                    })
                })
            }),
            "conditions": models.OktaSignOnPolicyRuleConditions({
                "authContext": models.PolicyRuleAuthContextCondition({
                    "authType": "ANY"
                })
            })
        })
        created_policy_rule, _, err = await client.create_policy_rule(
            created_policy.id, policy_rule_model
        )
        assert err is None
        assert isinstance(created_policy_rule, models.OktaSignOnPolicyRule)
        assert created_policy_rule.name == policy_rule_model.name
        assert created_policy_rule.type == "SIGN_ON"
        assert created_policy_rule.actions.signon.access == "ALLOW"
        assert created_policy_rule.actions.signon.require_factor is False
        assert created_policy_rule.actions.signon.remember_device_by_default\
            is False
        assert created_policy_rule.actions.signon.session.\
            use_persistent_cookie is False
        assert created_policy_rule.actions.signon.session.\
            max_session_idle_minutes == 720
        assert created_policy_rule.actions.signon.session.\
            max_session_lifetime_minutes == 0
        assert created_policy_rule.conditions.auth_context.auth_type == "ANY"

        # Retrieve
        retrieved_policy_rule, _, err = await client.get_policy_rule(
            created_policy.id, created_policy_rule.id)
        assert err is None
        assert retrieved_policy_rule.id == created_policy_rule.id
        assert retrieved_policy_rule.name == created_policy_rule.name
        assert isinstance(retrieved_policy_rule, models.OktaSignOnPolicyRule)

        # Update
        created_policy_rule.name = policy_rule_model.name + "UPDATED"
        created_policy_rule.conditions.network.connection = "ANYWHERE"
        updated_policy_rule, _, err = await client.update_policy_rule(
            created_policy.id, created_policy_rule.id, created_policy_rule
        )
        assert err is None
        assert isinstance(updated_policy_rule, models.OktaSignOnPolicyRule)
        assert updated_policy_rule.name == created_policy_rule.name
        assert updated_policy_rule.conditions.network.connection ==\
            created_policy_rule.conditions.network.connection

        # Retrieve
        retrieved_policy_rule, _, err = await client.get_policy_rule(
            created_policy.id, created_policy_rule.id)
        assert err is None
        assert retrieved_policy_rule.id == created_policy_rule.id
        assert retrieved_policy_rule.name == created_policy_rule.name
        assert retrieved_policy_rule.conditions.network.connection ==\
            created_policy_rule.conditions.network.connection
        assert isinstance(retrieved_policy_rule, models.OktaSignOnPolicyRule)

        # Delete
        _, err = await \
            client.delete_policy_rule(
                created_policy.id, created_policy_rule.id)
        assert err is None
        _, err = await client.delete_policy(created_policy.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_list_policy_rules(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Policy
        policy_model = models.OktaSignOnPolicy({
            "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Sign-On",
            "status": "ACTIVE",
            "description": "Test policy applies for tests only"
        })

        created_policy, _, err = await client.create_policy(policy_model)
        assert err is None
        assert isinstance(created_policy, models.OktaSignOnPolicy)
        assert created_policy.name == policy_model.name
        assert created_policy.description == policy_model.description
        assert created_policy.status == "ACTIVE"
        assert created_policy.type == models.PolicyType.OKTA_SIGN_ON

        # Create Policy Rule
        policy_rule_model = models.OktaSignOnPolicyRule({
            "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Sign-On-Rule",
            "actions": models.OktaSignOnPolicyRuleActions({
                "signon": models.OktaSignOnPolicyRuleSignonActions({
                    "access": "ALLOW",
                    "requireFactor": "false",
                    "rememberDeviceByDefault": "false",
                    "session":
                    models.OktaSignOnPolicyRuleSignonSessionActions({
                        "usePersistentCookie": "false",
                        "maxSessionIdleMinutes": 720,
                        "maxSessionLifetimeMinutes": 0
                    })
                })
            }),
            "conditions": models.OktaSignOnPolicyRuleConditions({
                "authContext": models.PolicyRuleAuthContextCondition({
                    "authType": "ANY"
                })
            })
        })
        created_policy_rule, _, err = await client.create_policy_rule(
            created_policy.id, policy_rule_model
        )
        assert err is None
        assert isinstance(created_policy_rule, models.OktaSignOnPolicyRule)
        assert created_policy_rule.name == policy_rule_model.name
        assert created_policy_rule.type == "SIGN_ON"
        assert created_policy_rule.actions.signon.access == "ALLOW"
        assert created_policy_rule.actions.signon.require_factor is False
        assert created_policy_rule.actions.signon.remember_device_by_default\
            is False
        assert created_policy_rule.actions.signon.session.\
            use_persistent_cookie is False
        assert created_policy_rule.actions.signon.session.\
            max_session_idle_minutes == 720
        assert created_policy_rule.actions.signon.session.\
            max_session_lifetime_minutes == 0
        assert created_policy_rule.conditions.auth_context.auth_type == "ANY"

        # List
        policy_rules, _, err = await \
            client.list_policy_rules(created_policy.id)
        assert err is None
        assert isinstance(policy_rules, list)
        assert next((plcy_rule for plcy_rule in policy_rules
                     if plcy_rule.id == created_policy_rule.id))
        assert len(policy_rules) == 1

        # Delete
        _, err = await \
            client.delete_policy_rule(
                created_policy.id, created_policy_rule.id)
        assert err is None
        _, err = await client.delete_policy(created_policy.id)
        assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_delete_policy_rules(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Policy
        policy_model = models.OktaSignOnPolicy({
            "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Sign-On",
            "status": "ACTIVE",
            "description": "Test policy applies for tests only"
        })

        created_policy, _, err = await client.create_policy(policy_model)
        assert err is None
        assert isinstance(created_policy, models.OktaSignOnPolicy)
        assert created_policy.name == policy_model.name
        assert created_policy.description == policy_model.description
        assert created_policy.status == "ACTIVE"
        assert created_policy.type == models.PolicyType.OKTA_SIGN_ON

        # Create Policy Rule
        policy_rule_model = models.OktaSignOnPolicyRule({
            "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Sign-On-Rule",
            "actions": models.OktaSignOnPolicyRuleActions({
                "signon": models.OktaSignOnPolicyRuleSignonActions({
                    "access": "ALLOW",
                    "requireFactor": "false",
                    "rememberDeviceByDefault": "false",
                    "session":
                    models.OktaSignOnPolicyRuleSignonSessionActions({
                        "usePersistentCookie": "false",
                        "maxSessionIdleMinutes": 720,
                        "maxSessionLifetimeMinutes": 0
                    })
                })
            }),
            "conditions": models.OktaSignOnPolicyRuleConditions({
                "authContext": models.PolicyRuleAuthContextCondition({
                    "authType": "ANY"
                })
            })
        })
        created_policy_rule, _, err = await client.create_policy_rule(
            created_policy.id, policy_rule_model
        )
        assert err is None
        assert isinstance(created_policy_rule, models.OktaSignOnPolicyRule)
        assert created_policy_rule.name == policy_rule_model.name
        assert created_policy_rule.type == "SIGN_ON"
        assert created_policy_rule.actions.signon.access == "ALLOW"
        assert created_policy_rule.actions.signon.require_factor is False
        assert created_policy_rule.actions.signon.remember_device_by_default\
            is False
        assert created_policy_rule.actions.signon.session.\
            use_persistent_cookie is False
        assert created_policy_rule.actions.signon.session.\
            max_session_idle_minutes == 720
        assert created_policy_rule.actions.signon.session.\
            max_session_lifetime_minutes == 0
        assert created_policy_rule.conditions.auth_context.auth_type == "ANY"

        # Retrieve
        retrieved_policy_rule, _, err = await client.get_policy_rule(
            created_policy.id, created_policy_rule.id)
        assert err is None
        assert retrieved_policy_rule.id == created_policy_rule.id
        assert retrieved_policy_rule.name == created_policy_rule.name
        assert isinstance(retrieved_policy_rule, models.OktaSignOnPolicyRule)

        # Delete
        _, err = await \
            client.delete_policy_rule(
                created_policy.id, created_policy_rule.id)
        assert err is None
        _, err = await client.delete_policy(created_policy.id)
        assert err is None

        # Retrieve
        retrieved_policy_rule, resp, err = await client.get_policy_rule(
            created_policy.id, created_policy_rule.id)
        assert err is not None
        assert isinstance(err, OktaAPIError)
        assert resp.get_status() == HTTPStatus.NOT_FOUND
        assert retrieved_policy_rule is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_activate_deactivate_policy_rule(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Policy
        policy_model = models.OktaSignOnPolicy({
            "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Sign-On",
            "status": "ACTIVE",
            "description": "Test policy applies for tests only"
        })

        created_policy, _, err = await client.create_policy(policy_model)
        assert err is None
        assert isinstance(created_policy, models.OktaSignOnPolicy)
        assert created_policy.name == policy_model.name
        assert created_policy.description == policy_model.description
        assert created_policy.status == "ACTIVE"
        assert created_policy.type == models.PolicyType.OKTA_SIGN_ON

        # Create Policy Rule
        policy_rule_model = models.OktaSignOnPolicyRule({
            "name": f"{TestPoliciesResource.SDK_PREFIX} Test-Sign-On-Rule",
            "actions": models.OktaSignOnPolicyRuleActions({
                "signon": models.OktaSignOnPolicyRuleSignonActions({
                    "access": "ALLOW",
                    "requireFactor": "false",
                    "rememberDeviceByDefault": "false",
                    "session":
                    models.OktaSignOnPolicyRuleSignonSessionActions({
                        "usePersistentCookie": "false",
                        "maxSessionIdleMinutes": 720,
                        "maxSessionLifetimeMinutes": 0
                    })
                })
            }),
            "conditions": models.OktaSignOnPolicyRuleConditions({
                "authContext": models.PolicyRuleAuthContextCondition({
                    "authType": "ANY"
                })
            })
        })
        created_policy_rule, _, err = await client.create_policy_rule(
            created_policy.id, policy_rule_model
        )
        assert err is None
        assert isinstance(created_policy_rule, models.OktaSignOnPolicyRule)
        assert created_policy_rule.name == policy_rule_model.name
        assert created_policy_rule.type == "SIGN_ON"
        assert created_policy_rule.status == "ACTIVE"
        assert created_policy_rule.actions.signon.access == "ALLOW"
        assert created_policy_rule.actions.signon.require_factor is False
        assert created_policy_rule.actions.signon.remember_device_by_default\
            is False
        assert created_policy_rule.actions.signon.session.\
            use_persistent_cookie is False
        assert created_policy_rule.actions.signon.session.\
            max_session_idle_minutes == 720
        assert created_policy_rule.actions.signon.session.\
            max_session_lifetime_minutes == 0
        assert created_policy_rule.conditions.auth_context.auth_type == "ANY"

        # Retrieve
        retrieved_policy_rule, _, err = await client.get_policy_rule(
            created_policy.id, created_policy_rule.id)
        assert err is None
        assert retrieved_policy_rule.id == created_policy_rule.id
        assert retrieved_policy_rule.name == created_policy_rule.name
        assert retrieved_policy_rule.status == "ACTIVE"
        assert isinstance(retrieved_policy_rule, models.OktaSignOnPolicyRule)

        # Deactivate
        _, err = await client.deactivate_policy_rule(
            created_policy.id, created_policy_rule.id
        )
        assert err is None

        # Retrieve
        retrieved_policy_rule, _, err = await client.get_policy_rule(
            created_policy.id, created_policy_rule.id)
        assert err is None
        assert retrieved_policy_rule.id == created_policy_rule.id
        assert retrieved_policy_rule.name == created_policy_rule.name
        assert retrieved_policy_rule.status == "INACTIVE"
        assert isinstance(retrieved_policy_rule, models.OktaSignOnPolicyRule)

        # Reactivate
        _, err = await client.activate_policy_rule(
            created_policy.id, created_policy_rule.id
        )
        assert err is None

        # Retrieve
        retrieved_policy_rule, _, err = await client.get_policy_rule(
            created_policy.id, created_policy_rule.id)
        assert err is None
        assert retrieved_policy_rule.id == created_policy_rule.id
        assert retrieved_policy_rule.name == created_policy_rule.name
        assert retrieved_policy_rule.status == "ACTIVE"
        assert isinstance(retrieved_policy_rule, models.OktaSignOnPolicyRule)

        # Delete
        _, err = await \
            client.delete_policy_rule(
                created_policy.id, created_policy_rule.id)
        assert err is None
        _, err = await client.delete_policy(created_policy.id)
        assert err is None



import pytest
from tests.mocks import MockOktaClient
import okta.models as models
from http import HTTPStatus
from okta.errors.okta_api_error import OktaAPIError


class TestLinkedObjectsResource:
    """
    Integration Tests for the Linked Objects Resource
    """
    SDK_PREFIX = "python_sdk"

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_add_get_linked_object(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Add Linked Object definition
        linked_object_model = models.LinkedObject({
            "primary": models.LinkedObjectDetails({
                "name": f"{TestLinkedObjectsResource.SDK_PREFIX}_primary_test",
                "title": "Primary",
                "description": "Primary Link Property",
                "type": models.LinkedObjectDetailsType.USER
            }),
            "associated": models.LinkedObjectDetails({
                "name": f"{TestLinkedObjectsResource.SDK_PREFIX}_assoc_test",
                "title": "Associated",
                "description": "Associated Link Property",
                "type": models.LinkedObjectDetailsType.USER
            })
        })

        created_linked_object_definition, _, err = await client.\
            add_linked_object_definition(linked_object_model)
        assert err is None
        assert isinstance(created_linked_object_definition,
                          models.LinkedObject)
        assert created_linked_object_definition.primary
        assert created_linked_object_definition.associated

        # Retrieve by Primary Name
        retrieved_linked_object_definition, _, err = await \
            client.get_linked_object_definition(
                linked_object_model.primary.name)
        assert err is None
        assert isinstance(retrieved_linked_object_definition,
                          models.LinkedObject)
        assert retrieved_linked_object_definition.primary.name ==\
            created_linked_object_definition.primary.name
        assert retrieved_linked_object_definition.associated.name ==\
            created_linked_object_definition.associated.name
        assert retrieved_linked_object_definition.primary.title ==\
            created_linked_object_definition.primary.title
        assert retrieved_linked_object_definition.associated.title ==\
            created_linked_object_definition.associated.title
        assert retrieved_linked_object_definition.primary.type ==\
            created_linked_object_definition.primary.type
        assert retrieved_linked_object_definition.associated.type ==\
            created_linked_object_definition.associated.type

        # Retrieve by Associated Name
        retrieved_linked_object_definition, _, err = await \
            client.get_linked_object_definition(
                linked_object_model.associated.name)
        assert err is None
        assert isinstance(retrieved_linked_object_definition,
                          models.LinkedObject)
        assert retrieved_linked_object_definition.primary.name ==\
            created_linked_object_definition.primary.name
        assert retrieved_linked_object_definition.associated.name ==\
            created_linked_object_definition.associated.name
        assert retrieved_linked_object_definition.primary.title ==\
            created_linked_object_definition.primary.title
        assert retrieved_linked_object_definition.associated.title ==\
            created_linked_object_definition.associated.title
        assert retrieved_linked_object_definition.primary.type ==\
            created_linked_object_definition.primary.type
        assert retrieved_linked_object_definition.associated.type ==\
            created_linked_object_definition.associated.type

        # Delete Linked Object definition
        _, err = await \
            client.delete_linked_object_definition(
                linked_object_model.primary.name)

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_get_all_linked_objects(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Add Linked Object definition
        linked_object_model_1 = models.LinkedObject({
            "primary": models.LinkedObjectDetails({
                "name": f"{TestLinkedObjectsResource.SDK_PREFIX}_primary_t1",
                "title": "Primary",
                "description": "Primary Link Property",
                "type": models.LinkedObjectDetailsType.USER
            }),
            "associated": models.LinkedObjectDetails({
                "name": f"{TestLinkedObjectsResource.SDK_PREFIX}_assoc_t1",
                "title": "Associated",
                "description": "Associated Link Property",
                "type": models.LinkedObjectDetailsType.USER
            })
        })

        linked_object_model_2 = models.LinkedObject({
            "primary": models.LinkedObjectDetails({
                "name": f"{TestLinkedObjectsResource.SDK_PREFIX}_primary_t2",
                "title": "Primary",
                "description": "Primary Link Property",
                "type": models.LinkedObjectDetailsType.USER
            }),
            "associated": models.LinkedObjectDetails({
                "name": f"{TestLinkedObjectsResource.SDK_PREFIX}_assoc_t2",
                "title": "Associated",
                "description": "Associated Link Property",
                "type": models.LinkedObjectDetailsType.USER
            })
        })

        created_linked_object_definition, _, err = await client.\
            add_linked_object_definition(linked_object_model_1)
        assert err is None
        assert isinstance(created_linked_object_definition,
                          models.LinkedObject)
        assert created_linked_object_definition.primary
        assert created_linked_object_definition.associated

        created_linked_object_definition_2, _, err = await client.\
            add_linked_object_definition(linked_object_model_2)
        assert err is None
        assert isinstance(created_linked_object_definition_2,
                          models.LinkedObject)
        assert created_linked_object_definition_2.primary
        assert created_linked_object_definition_2.associated

        # List
        all_linked_obj_defs, _, err = await\
            client.list_linked_object_definitions()
        assert err is None
        assert len(all_linked_obj_defs) > 0
        assert next((lo for lo in all_linked_obj_defs
                     if linked_object_model_1.primary.name == lo.primary.name))
        assert next((lo for lo in all_linked_obj_defs
                     if linked_object_model_2.primary.name == lo.primary.name))

        # Delete Linked Object definition
        _, err = await \
            client.delete_linked_object_definition(
                linked_object_model_1.primary.name)
        _, err = await \
            client.delete_linked_object_definition(
                linked_object_model_2.primary.name)

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_delete_linked_object(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Add Linked Object definition
        linked_object_model = models.LinkedObject({
            "primary": models.LinkedObjectDetails({
                "name": f"{TestLinkedObjectsResource.SDK_PREFIX}_primary_test",
                "title": "Primary",
                "description": "Primary Link Property",
                "type": models.LinkedObjectDetailsType.USER
            }),
            "associated": models.LinkedObjectDetails({
                "name": f"{TestLinkedObjectsResource.SDK_PREFIX}_assoc_test",
                "title": "Associated",
                "description": "Associated Link Property",
                "type": models.LinkedObjectDetailsType.USER
            })
        })

        created_linked_object_definition, _, err = await client.\
            add_linked_object_definition(linked_object_model)
        assert err is None
        assert isinstance(created_linked_object_definition,
                          models.LinkedObject)
        assert created_linked_object_definition.primary
        assert created_linked_object_definition.associated

        # Retrieve by Primary Name
        retrieved_linked_object_definition, _, err = await \
            client.get_linked_object_definition(
                linked_object_model.primary.name)
        assert err is None
        assert isinstance(retrieved_linked_object_definition,
                          models.LinkedObject)
        assert retrieved_linked_object_definition.primary.name ==\
            created_linked_object_definition.primary.name
        assert retrieved_linked_object_definition.associated.name ==\
            created_linked_object_definition.associated.name
        assert retrieved_linked_object_definition.primary.title ==\
            created_linked_object_definition.primary.title
        assert retrieved_linked_object_definition.associated.title ==\
            created_linked_object_definition.associated.title
        assert retrieved_linked_object_definition.primary.type ==\
            created_linked_object_definition.primary.type
        assert retrieved_linked_object_definition.associated.type ==\
            created_linked_object_definition.associated.type

        # Delete Linked Object definition
        _, err = await \
            client.delete_linked_object_definition(
                linked_object_model.primary.name)

        # Retrieve by Primary Name
        retrieved_linked_object_definition, resp, err = await \
            client.get_linked_object_definition(
                linked_object_model.primary.name)
        assert err is not None
        assert isinstance(err, OktaAPIError)
        assert resp.get_status() == HTTPStatus.NOT_FOUND
        assert retrieved_linked_object_definition is None



"""
Copyright 2020 - Present Okta, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# AUTO-GENERATED! DO NOT EDIT FILE DIRECTLY
# SEE CONTRIBUTOR DOCUMENTATION

from urllib.parse import urlencode
from okta.models.trusted_origin\
    import TrustedOrigin
from okta.utils import format_url
from okta.api_client import APIClient


class TrustedOriginClient(APIClient):
    """
    A Client object for the TrustedOrigin resource.
    """

    def __init__(self):
        self._base_url = ""

    async def list_origins(
            self, query_params={}
    ):
        """
        Args:
            query_params {dict}: Map of query parameters for request
            [query_params.q] {str}
            [query_params.filter] {str}
            [query_params.after] {str}
            [query_params.limit] {str}
        Returns:
            list: Collection of TrustedOrigin instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/trustedOrigins
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, TrustedOrigin)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(TrustedOrigin(
                    self.form_response_body(item)
                    ))
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def create_origin(
            self, trusted_origin
    ):
        """
        Args:
            {trusted_origin}
        Returns:
            TrustedOrigin
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/trustedOrigins
            """)

        if isinstance(trusted_origin, dict):
            body = trusted_origin
        else:
            body = trusted_origin.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, TrustedOrigin)

        if error:
            return (None, response, error)

        try:
            result = TrustedOrigin(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def delete_origin(
            self, trustedOriginId
    ):
        """
        Args:
            trusted_origin_id {str}
        """
        http_method = "delete".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/trustedOrigins/{trustedOriginId}
            """)

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def get_origin(
            self, trustedOriginId
    ):
        """
        Args:
            trusted_origin_id {str}
        Returns:
            TrustedOrigin
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/trustedOrigins/{trustedOriginId}
            """)

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, TrustedOrigin)

        if error:
            return (None, response, error)

        try:
            result = TrustedOrigin(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def update_origin(
            self, trustedOriginId, trusted_origin
    ):
        """
        Args:
            trusted_origin_id {str}
            {trusted_origin}
        Returns:
            TrustedOrigin
        """
        http_method = "put".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/trustedOrigins/{trustedOriginId}
            """)

        if isinstance(trusted_origin, dict):
            body = trusted_origin
        else:
            body = trusted_origin.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, TrustedOrigin)

        if error:
            return (None, response, error)

        try:
            result = TrustedOrigin(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def activate_origin(
            self, trustedOriginId
    ):
        """
        Args:
            trusted_origin_id {str}
        Returns:
            TrustedOrigin
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/trustedOrigins/{trustedOriginId}/lifecycle
                /activate
            """)

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, TrustedOrigin)

        if error:
            return (None, response, error)

        try:
            result = TrustedOrigin(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def deactivate_origin(
            self, trustedOriginId
    ):
        """
        Args:
            trusted_origin_id {str}
        Returns:
            TrustedOrigin
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/trustedOrigins/{trustedOriginId}/lifecycle
                /deactivate
            """)

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, TrustedOrigin)

        if error:
            return (None, response, error)

        try:
            result = TrustedOrigin(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)



import aiohttp
import asyncio
import logging
from aiohttp.client_reqrep import ConnectionKey
from ssl import SSLCertVerificationError

from okta.client import Client as OktaClient
import pytest
from okta.constants import FINDING_OKTA_DOMAIN
import yaml
import os
from okta.error_messages import ERROR_MESSAGE_API_TOKEN_DEFAULT, \
    ERROR_MESSAGE_API_TOKEN_MISSING, ERROR_MESSAGE_AUTH_MODE_INVALID, \
    ERROR_MESSAGE_CLIENT_ID_DEFAULT, ERROR_MESSAGE_CLIENT_ID_MISSING,\
    ERROR_MESSAGE_ORG_URL_ADMIN, ERROR_MESSAGE_ORG_URL_MISSING, \
    ERROR_MESSAGE_ORG_URL_NOT_HTTPS, ERROR_MESSAGE_ORG_URL_TYPO, \
    ERROR_MESSAGE_ORG_URL_YOUROKTADOMAIN, ERROR_MESSAGE_SCOPES_PK_MISSING, \
    ERROR_MESSAGE_PROXY_MISSING_HOST, ERROR_MESSAGE_PROXY_MISSING_AUTH, \
    ERROR_MESSAGE_PROXY_INVALID_PORT
from okta.constants import _GLOBAL_YAML_PATH, _LOCAL_YAML_PATH
from okta.exceptions import HTTPException
from okta.http_client import HTTPClient


"""
Testing Okta Client Instantiation in different scenarios
"""


def test_constructor_user_config_empty(fs):
    config = {}
    with pytest.raises(ValueError) as exception_info:
        OktaClient(user_config=config)
    assert ERROR_MESSAGE_ORG_URL_MISSING in str(exception_info.value)
    assert ERROR_MESSAGE_API_TOKEN_MISSING in str(exception_info.value)


def test_constructor_user_config_url_empty():
    config = {'orgUrl': '', 'token': 'TOKEN'}
    with pytest.raises(ValueError) as exception_info:
        OktaClient(user_config=config)
    assert ERROR_MESSAGE_ORG_URL_MISSING in str(exception_info.value)


def test_constructor_user_config_url_not_https():
    config = {'orgUrl': 'http://test.okta.com', 'token': 'TOKEN'}
    with pytest.raises(ValueError) as exception_info:
        OktaClient(user_config=config)
    assert ERROR_MESSAGE_ORG_URL_NOT_HTTPS in str(exception_info.value)
    assert FINDING_OKTA_DOMAIN in str(exception_info.value)


def test_constructor_user_config_url_has_yourOktaDomain():
    config = {
        'orgUrl': 'https://{yourOktaDomain}.okta.com', 'token': 'TOKEN'
    }
    with pytest.raises(ValueError) as exception_info:
        OktaClient(user_config=config)
    assert ERROR_MESSAGE_ORG_URL_YOUROKTADOMAIN in str(exception_info.value)


@ pytest.mark.parametrize("url", ["https://dev-admin.okta.com",
                                  "https://dev-admin.oktapreview.com",
                                  "https://dev-admin.okta-emea.com",
                                  "https://test-admin.okta.com"])
def test_constructor_user_config_url_has_admin(url):
    config = {
        'orgUrl': url, 'token': 'TOKEN'
    }
    with pytest.raises(ValueError) as exception_info:
        OktaClient(user_config=config)
    assert all(string in str(exception_info.value) for string in [
               ERROR_MESSAGE_ORG_URL_ADMIN, f"Current value: {url}"])


def test_constructor_user_config_url_dot_com_twice():
    url = 'https://test.okta.com.com'
    config = {
        'orgUrl': url, 'token': 'TOKEN'
    }
    with pytest.raises(ValueError) as exception_info:
        OktaClient(user_config=config)
    assert all(string in str(exception_info.value) for string in [
               ERROR_MESSAGE_ORG_URL_TYPO, f"Current value: {url}"])


def test_constructor_user_config_url_punctuation():
    # test for urls with '://' multiple times
    url = 'https://://test.okta.com'
    config = {
        'orgUrl': url, 'token': 'TOKEN'
    }
    with pytest.raises(ValueError) as exception_info:
        OktaClient(user_config=config)
    assert all(string in str(exception_info.value) for string in [
               ERROR_MESSAGE_ORG_URL_TYPO, f"Current value: {url}"])


def test_constructor_user_config_token_empty(fs):
    config = {'orgUrl': 'https://test.okta.com', 'token': ''}
    with pytest.raises(ValueError) as exception_info:
        OktaClient(user_config=config)
    assert ERROR_MESSAGE_API_TOKEN_MISSING in str(exception_info.value)


def test_constructor_user_config_url_has_apiToken(fs):
    config = {
        'orgUrl': 'https://test.okta.com', 'token': '{apiToken}'
    }
    with pytest.raises(ValueError) as exception_info:
        OktaClient(user_config=config)
    assert ERROR_MESSAGE_API_TOKEN_DEFAULT in str(exception_info.value)


def test_constructor_user_config_auth_mode_invalid():
    authorizationMode = "blah"
    config = {'orgUrl': "https://test.okta.com",
              'token': "TOKEN",
              'authorizationMode': authorizationMode}
    with pytest.raises(ValueError) as exception_info:
        OktaClient(user_config=config)
    assert all(string in str(exception_info.value) for string in [
               ERROR_MESSAGE_AUTH_MODE_INVALID, f"with {authorizationMode}"])


def test_constructor_user_config_SSWS():
    org_url = "https://test.okta.com"
    token = "TOKEN"
    config = {'orgUrl': org_url, 'token': token}
    client = OktaClient(user_config=config)
    loaded_config = client.get_config()
    assert org_url == loaded_config['client']['orgUrl']
    assert token == loaded_config['client']['token']


@ pytest.mark.parametrize("private_key", ["private key hash",
                                          "pem_file.pem",
                                          "{'Jwks'}"])
def test_constructor_user_config_PK(private_key):
    org_url = "https://test.okta.com"
    authorizationMode = "PrivateKey"
    client_id = "clientID"
    scopes = ["scope1"]

    config = {
        'orgUrl': org_url,
        'authorizationMode': authorizationMode,
        'clientId': client_id,
        'scopes': scopes,
        'privateKey': private_key
    }
    client = OktaClient(user_config=config)
    loaded_config = client.get_config()
    assert org_url == loaded_config['client']['orgUrl']
    assert authorizationMode == loaded_config['client']['authorizationMode']
    assert client_id == loaded_config['client']['clientId']
    assert scopes == loaded_config['client']['scopes']
    assert private_key == loaded_config['client']['privateKey']


def test_constructor_user_config_PK_empty(fs):
    org_url = "https://test.okta.com"
    authorizationMode = "PrivateKey"

    config = {
        'orgUrl': org_url,
        'authorizationMode': authorizationMode,
    }
    with pytest.raises(ValueError) as exception_info:
        OktaClient(user_config=config)
    assert all(string in str(exception_info.value) for string in [
               ERROR_MESSAGE_CLIENT_ID_MISSING, ERROR_MESSAGE_SCOPES_PK_MISSING
               ])


def test_constructor_user_config_PK_client_id_empty():
    org_url = "https://test.okta.com"
    authorizationMode = "PrivateKey"
    scopes = ["scope1"]
    private_key_hash = "private key hash"

    config = {
        'orgUrl': org_url,
        'authorizationMode': authorizationMode,
        'clientId': "",
        'scopes': scopes,
        'privateKey': private_key_hash
    }
    with pytest.raises(ValueError) as exception_info:
        OktaClient(user_config=config)
    assert all(string in str(exception_info.value) for string in [
               ERROR_MESSAGE_CLIENT_ID_MISSING
               ])


def test_constructor_user_config_PK_client_id_default():
    org_url = "https://test.okta.com"
    authorizationMode = "PrivateKey"
    scopes = ["scope1"]
    private_key_hash = "private key hash"

    config = {
        'orgUrl': org_url,
        'authorizationMode': authorizationMode,
        'clientId': "{clientId}",
        'scopes': scopes,
        'privateKey': private_key_hash
    }
    with pytest.raises(ValueError) as exception_info:
        OktaClient(user_config=config)
    assert all(string in str(exception_info.value) for string in [
               ERROR_MESSAGE_CLIENT_ID_DEFAULT
               ])


@ pytest.mark.parametrize("scopes,private_key", [([], "private key hash"),
                                                 (["scope1"], ""),
                                                 ([], "")])
def test_constructor_user_config_PK_scopes_and_or_private_key_empty(
        scopes,
        private_key):
    org_url = "https://test.okta.com"
    authorizationMode = "PrivateKey"
    client_id = "clientID"

    config = {
        'orgUrl': org_url,
        'authorizationMode': authorizationMode,
        'clientId': client_id,
        'scopes': scopes,
        'privateKey': private_key
    }
    with pytest.raises(ValueError) as exception_info:
        OktaClient(user_config=config)
    assert all(string in str(exception_info.value) for string in [
               ERROR_MESSAGE_SCOPES_PK_MISSING
               ])


"""
Testing constructor with YAML configurations
"""


def test_constructor_global_config_SSWS(fs):
    fs.pause()
    global_sample = os.path.join(os.path.dirname(
        __file__), "files", "SSWS-sample-global.yaml")
    with open(global_sample) as file:
        global_config = yaml.load(file, Loader=yaml.SafeLoader)
        org_url = global_config["okta"]["client"]["orgUrl"]
        token = global_config["okta"]["client"]["token"]
        fs.resume()
        fs.create_file(_GLOBAL_YAML_PATH, contents=yaml.dump(global_config))

    client = OktaClient()
    loaded_config = client.get_config()

    assert org_url == loaded_config['client']['orgUrl']
    assert token == loaded_config['client']['token']


def test_constructor_local_config_SSWS(fs):
    fs.pause()
    local_sample = os.path.join(os.path.dirname(
        __file__), "files", "SSWS-sample-local.yaml")
    with open(local_sample) as file:
        local_config = yaml.load(file, Loader=yaml.SafeLoader)
        org_url = local_config["okta"]["client"]["orgUrl"]
        token = local_config["okta"]["client"]["token"]
        fs.resume()
        fs.create_file(_LOCAL_YAML_PATH, contents=yaml.dump(local_config))

    client = OktaClient()
    loaded_config = client.get_config()

    assert org_url == loaded_config['client']['orgUrl']
    assert token == loaded_config['client']['token']


def test_constructor_global_config_PK(fs):
    fs.pause()
    global_sample = os.path.join(os.path.dirname(
        __file__), "files", "PK-sample-global.yaml")
    with open(global_sample) as file:
        global_config = yaml.load(file, Loader=yaml.SafeLoader)
        org_url = global_config["okta"]["client"]["orgUrl"]
        client_id = global_config["okta"]["client"]["clientId"]
        private_key = global_config["okta"]["client"]["privateKey"]
        fs.resume()
        fs.create_file(_GLOBAL_YAML_PATH, contents=yaml.dump(global_config))

    client = OktaClient()
    loaded_config = client.get_config()

    assert org_url == loaded_config['client']['orgUrl']
    assert client_id == loaded_config['client']['clientId']
    assert private_key == loaded_config['client']['privateKey']


def test_constructor_local_config_PK(fs):
    fs.pause()
    local_sample = os.path.join(os.path.dirname(
        __file__), "files", "PK-sample-local.yaml")
    with open(local_sample) as file:
        local_config = yaml.load(file, Loader=yaml.SafeLoader)
        org_url = local_config["okta"]["client"]["orgUrl"]
        client_id = local_config["okta"]["client"]["clientId"]
        private_key = local_config["okta"]["client"]["privateKey"]
        fs.resume()
        fs.create_file(_LOCAL_YAML_PATH, contents=yaml.dump(local_config))

    client = OktaClient()
    loaded_config = client.get_config()

    assert org_url == loaded_config['client']['orgUrl']
    assert client_id == loaded_config['client']['clientId']
    assert private_key == loaded_config['client']['privateKey']


def test_constructor_env_vars_SSWS():
    org_url = "https://test.okta.com"
    token = "TOKEN"
    os.environ["OKTA_CLIENT_ORGURL"] = org_url
    os.environ["OKTA_CLIENT_TOKEN"] = token

    client = OktaClient()
    loaded_config = client.get_config()

    os.environ.pop("OKTA_CLIENT_ORGURL")
    os.environ.pop("OKTA_CLIENT_TOKEN")

    assert org_url == loaded_config['client']['orgUrl']
    assert token == loaded_config['client']['token']


def test_constructor_env_vars_PK():
    authorizationMode = "PrivateKey"
    org_url = "https://test.okta.com"
    client_id = "clientID"
    scopes = "scope1,scope2,scope3"
    private_key = "private key"
    os.environ["OKTA_CLIENT_AUTHORIZATIONMODE"] = authorizationMode
    os.environ["OKTA_CLIENT_ORGURL"] = org_url
    os.environ["OKTA_CLIENT_CLIENTID"] = client_id
    os.environ["OKTA_CLIENT_SCOPES"] = scopes
    os.environ["OKTA_CLIENT_PRIVATEKEY"] = private_key

    client = OktaClient()
    loaded_config = client.get_config()

    os.environ.pop("OKTA_CLIENT_ORGURL")
    os.environ.pop("OKTA_CLIENT_AUTHORIZATIONMODE")
    os.environ.pop("OKTA_CLIENT_CLIENTID")
    os.environ.pop("OKTA_CLIENT_SCOPES")
    os.environ.pop("OKTA_CLIENT_PRIVATEKEY")

    assert authorizationMode == loaded_config['client']['authorizationMode']
    assert org_url == loaded_config['client']['orgUrl']
    assert client_id == loaded_config['client']['clientId']
    assert scopes.split(',') == loaded_config['client']['scopes']
    assert private_key == loaded_config['client']['privateKey']


def test_constructor_precedence_highest_rank_local_yaml(fs):
    # Setup Global config
    fs.pause()
    global_sample = os.path.join(os.path.dirname(
        __file__), "files", "SSWS-sample-global.yaml")
    with open(global_sample) as file:
        global_config = yaml.load(file, Loader=yaml.SafeLoader)
        global_org_url = global_config["okta"]["client"]["orgUrl"]
        global_token = global_config["okta"]["client"]["token"]
        fs.resume()
        fs.create_file(_GLOBAL_YAML_PATH, contents=yaml.dump(global_config))

    # Setup Local config
    fs.pause()
    local_sample = os.path.join(os.path.dirname(
        __file__), "files", "SSWS-sample-local.yaml")
    with open(local_sample) as file:
        local_config = yaml.load(file, Loader=yaml.SafeLoader)
        local_org_url = local_config["okta"]["client"]["orgUrl"]
        local_token = local_config["okta"]["client"]["token"]
        fs.resume()
        fs.create_file(_LOCAL_YAML_PATH, contents=yaml.dump(local_config))

    # Create client and validate values
    client = OktaClient()
    loaded_config = client.get_config()

    assert local_org_url == loaded_config['client']['orgUrl']
    assert local_token == loaded_config['client']['token']
    assert local_org_url != global_org_url
    assert local_token != global_token
    assert global_org_url != loaded_config['client']['orgUrl']
    assert global_token != loaded_config['client']['token']


def test_constructor_precedence_highest_rank_env_vars(fs):
    # Setup Local config
    fs.pause()
    local_sample = os.path.join(os.path.dirname(
        __file__), "files", "SSWS-sample-local.yaml")
    with open(local_sample) as file:
        local_config = yaml.load(file, Loader=yaml.SafeLoader)
        local_org_url = local_config["okta"]["client"]["orgUrl"]
        local_token = local_config["okta"]["client"]["token"]
        fs.resume()
        fs.create_file(_LOCAL_YAML_PATH, contents=yaml.dump(local_config))
    # Setup env. vars
    env_org_url = "https://test.env.okta.com"
    env_token = "envTOKEN"
    os.environ["OKTA_CLIENT_ORGURL"] = env_org_url
    os.environ["OKTA_CLIENT_TOKEN"] = env_token

    client = OktaClient()
    loaded_config = client.get_config()

    os.environ.pop("OKTA_CLIENT_ORGURL")
    os.environ.pop("OKTA_CLIENT_TOKEN")

    assert local_org_url != loaded_config['client']['orgUrl']
    assert local_token != loaded_config['client']['token']
    assert local_org_url != env_org_url
    assert local_token != env_token
    assert env_org_url == loaded_config['client']['orgUrl']
    assert env_token == loaded_config['client']['token']


def test_constructor_precedence_highest_rank_user_config():
    # Setup env. vars
    env_org_url = "https://test.env.okta.com"
    env_token = "envTOKEN"
    os.environ["OKTA_CLIENT_ORGURL"] = env_org_url
    os.environ["OKTA_CLIENT_TOKEN"] = env_token

    # Setup user config
    user_org_url = "https://test.user.okta.com"
    user_token = "userTOKEN"
    config = {'orgUrl': user_org_url, 'token': user_token}

    client = OktaClient(config)
    loaded_config = client.get_config()

    os.environ.pop("OKTA_CLIENT_ORGURL")
    os.environ.pop("OKTA_CLIENT_TOKEN")

    assert user_org_url == loaded_config['client']['orgUrl']
    assert user_token == loaded_config['client']['token']
    assert user_org_url != env_org_url
    assert user_token != env_token
    assert env_org_url != loaded_config['client']['orgUrl']
    assert env_token != loaded_config['client']['token']


def test_constructor_valid_proxy():
    org_url = "https://test.okta.com"
    token = "TOKEN"

    port = 8080
    host = "test.okta.com"
    username = "username"
    password = "password"

    config = {
        'orgUrl': org_url,
        'token': token,
        'proxy': {
            'port': port,
            'host': host,
            'username': username,
            'password': password
        }
    }
    # Ensure no error is raised and correct proxy is determined
    client = OktaClient(user_config=config)
    assert client.get_request_executor(
    )._http_client._proxy == f"http://{username}:{password}@{host}:{port}/"


def test_constructor_valid_no_proxy():
    org_url = "https://test.okta.com"
    token = "TOKEN"

    config = {
        'orgUrl': org_url,
        'token': token
    }

    # Ensure no error is raised and proxy is None
    client = OktaClient(user_config=config)
    assert client.get_request_executor(
    )._http_client._proxy is None


def test_constructor_valid_env_vars():
    org_url = "https://test.okta.com"
    token = "TOKEN"

    config = {
        'orgUrl': org_url,
        'token': token
    }

    # Setting up env vars
    os.environ["HTTP_PROXY"] = "http://user:pass@test.okta.com:8080"
    os.environ["HTTPS_PROXY"] = "https://user:pass@test.okta.com:8080"

    expected = os.environ["HTTPS_PROXY"]
    client = OktaClient(user_config=config)

    # Deleting env vars
    del os.environ['HTTP_PROXY']
    del os.environ['HTTPS_PROXY']

    # Ensure no error is raised and proxy is None
    assert client.get_request_executor(
    )._http_client._proxy == expected


def test_constructor_invalid_missing_host():
    org_url = "https://test.okta.com"
    token = "TOKEN"

    port = 8080
    username = "username"
    password = "password"

    config = {
        'orgUrl': org_url,
        'token': token,
        'proxy': {
            'port': port,

            'username': username,
            'password': password
        }
    }

    # Expect error with config
    with pytest.raises(ValueError) as exception_info:
        OktaClient(user_config=config)
        assert ERROR_MESSAGE_PROXY_MISSING_HOST in exception_info.value


@pytest.mark.parametrize("username,password", [("", "password"),
                                               ("username", "")])
def test_constructor_invalid_missing_username_or_password(username, password):
    org_url = "https://test.okta.com"
    token = "TOKEN"

    port = 8080
    host = "test.okta.com"

    config = {
        'orgUrl': org_url,
        'token': token,
        'proxy': {
            'port': port,
            'host': host,
            'username': username,
            'password': password
        }
    }

    # Expect error with config
    with pytest.raises(ValueError) as exception_info:
        OktaClient(user_config=config)
        assert ERROR_MESSAGE_PROXY_MISSING_AUTH in exception_info.value


@pytest.mark.parametrize("port", [-1, 0, 65536, "port"])
def test_constructor_invalid_port_number(port):
    org_url = "https://test.okta.com"
    token = "TOKEN"

    host = "test.okta.com"
    username = "username"
    password = "password"

    config = {
        'orgUrl': org_url,
        'token': token,
        'proxy': {
            'port': port,
            'host': host,
            'username': username,
            'password': password
        }
    }

    # Expect error with config
    with pytest.raises(ValueError) as exception_info:
        OktaClient(user_config=config)
        assert ERROR_MESSAGE_PROXY_INVALID_PORT in exception_info.value


def test_constructor_custom_http_client_impl():
    class CustomHTTPClient(HTTPClient):
        pass
    config = {'httpClient': CustomHTTPClient}
    client = OktaClient(config)
    assert isinstance(client._request_executor._http_client, CustomHTTPClient)


def test_constructor_client_logging():
    logger = logging.getLogger('okta-sdk-python')
    assert logger.disabled
    config = {'logging': {"enabled": True, "logLevel": logging.DEBUG}}
    client = OktaClient(config)
    assert not logger.disabled
    assert logger.level == logging.DEBUG


def test_client_raise_exception():
    org_url = "https://test.okta.com"
    token = "TOKEN"
    config = {'orgUrl': org_url, 'token': token, 'raiseException': True}
    client = OktaClient(config)
    with pytest.raises(HTTPException):
        asyncio.run(client.list_users())


def test_client_custom_headers(monkeypatch, mocker):
    org_url = "https://test.okta.com"
    token = "TOKEN"
    config = {'orgUrl': org_url, 'token': token}
    custom_headers = {'Header-Test-1': 'test value 1',
                      'Header-Test-2': 'test value 2'}
    client = OktaClient(config)

    # verify custom headers are set
    client.set_custom_headers(custom_headers)
    assert client.get_custom_headers() == custom_headers

    # mock http requests, verify if custom header is present in request
    class MockHTTPRequest():
        def __call__(self, **params):
            self.request_info = params
            self.headers = params['headers']
            self.url = params['url']
            self.content_type = 'application/json'
            self.links = ''
            self.text = MockHTTPRequest.mock_response_text
            self.status = 200
            return self

        async def __aenter__(self):
            return self

        async def __aexit__(self, exc_type, exc, tb):
            pass

        @staticmethod
        async def mock_response_text():
            return '[{"text": "mock response text"}]'

    mock_http_request = MockHTTPRequest()
    monkeypatch.setattr(aiohttp, 'request', mock_http_request)
    asyncio.run(client.list_users())
    assert 'Header-Test-1' in mock_http_request.headers
    assert 'Header-Test-2' in mock_http_request.headers

    # verify custom headers are cleared
    client.clear_custom_headers()
    assert client.get_custom_headers() == {}


def test_client_handle_aiohttp_error(monkeypatch, mocker):
    org_url = "https://test.okta.com"
    token = "TOKEN"
    config = {'orgUrl': org_url, 'token': token}
    client = OktaClient(config)

    class MockHTTPRequest():
        def __call__(self, **params):
            raise aiohttp.ClientConnectorCertificateError(
                ConnectionKey(host=org_url,
                              port=443,
                              is_ssl=True,
                              ssl=None,
                              proxy=None,
                              proxy_auth=None,
                              proxy_headers_hash=None),
                SSLCertVerificationError(1,
                                         '[SSL: CERTIFICATE_VERIFY_FAILED] certificate verify failed: '
                                         'unable to get local issuer certificate (_ssl.c:1123)'))

        async def __aenter__(self):
            return self

        async def __aexit__(self, exc_type, exc, tb):
            pass

        @staticmethod
        async def mock_response_text():
            return '[{"text": "mock response text"}]'

    mock_http_request = MockHTTPRequest()
    monkeypatch.setattr(aiohttp, 'request', mock_http_request)
    res, resp_body, error = asyncio.run(client.list_users())
    assert res is None
    assert resp_body is None
    assert isinstance(error, aiohttp.ClientError)


def test_client_log_debug(monkeypatch, caplog):
    org_url = "https://test.okta.com"
    token = "TOKEN"
    config = {'orgUrl': org_url, 'token': token,
              'logging': {'enabled': True, 'logLevel': logging.DEBUG}}
    client = OktaClient(config)

    class MockHTTPRequest():
        def __call__(self, **params):
            self.request_info = params
            self.headers = params['headers']
            self.url = params['url']
            self.content_type = 'application/json'
            self.links = ''
            self.text = MockHTTPRequest.mock_response_text
            self.status = 200
            return self

        async def __aenter__(self):
            return self

        async def __aexit__(self, exc_type, exc, tb):
            pass

        @staticmethod
        async def mock_response_text():
            return '[{"embedded": null,' \
                   '"links": {"self": {"href": "https://test.okta.com/v1/users/test_id"}},' \
                   '"activated": "2021-01-01T00:00:00.000Z",' \
                   '"created": "2021-01-01T00:00:00.000Z",' \
                   '"credentials": null,' \
                   '"id": "test_id",' \
                   '"last_login": null,' \
                   '"profile": {"name": "test_name"},' \
                   '"status": null,' \
                   '"status_changed": null,' \
                   '"transitioning_to_status": null,' \
                   '"type": null}]'

    mock_http_request = MockHTTPRequest()
    monkeypatch.setattr(aiohttp, 'request', mock_http_request)
    with caplog.at_level(logging.DEBUG):
        res, resp_body, error = asyncio.run(client.list_users())
        assert 'okta-sdk-python' in caplog.text
        assert 'DEBUG' in caplog.text
        assert "'method': 'GET'" in caplog.text
        assert "'url': 'https://test.okta.com/api/v1/users'" in caplog.text


def test_client_log_info(monkeypatch, caplog):
    org_url = "https://test.okta.com"
    token = "TOKEN"
    config = {'orgUrl': org_url, 'token': token,
              'logging': {'enabled': True, 'logLevel': logging.INFO}}
    client = OktaClient(config)

    class MockHTTPRequest():
        def __call__(self, **params):
            self.request_info = params
            self.headers = params['headers']
            self.url = params['url']
            self.content_type = 'application/json'
            self.links = ''
            self.text = MockHTTPRequest.mock_response_text
            self.status = 200
            return self

        async def __aenter__(self):
            return self

        async def __aexit__(self, exc_type, exc, tb):
            pass

        @staticmethod
        async def mock_response_text():
            return '[{"embedded": null,' \
                   '"links": {"self": {"href": "https://test.okta.com/v1/users/test_id"}},' \
                   '"activated": "2021-01-01T00:00:00.000Z",' \
                   '"created": "2021-01-01T00:00:00.000Z",' \
                   '"credentials": null,' \
                   '"id": "test_id",' \
                   '"last_login": null,' \
                   '"profile": {"name": "test_name"},' \
                   '"status": null,' \
                   '"status_changed": null,' \
                   '"transitioning_to_status": null,' \
                   '"type": null}]'

    mock_http_request = MockHTTPRequest()
    monkeypatch.setattr(aiohttp, 'request', mock_http_request)
    with caplog.at_level(logging.INFO):
        res, resp_body, error = asyncio.run(client.list_users())
        assert caplog.text == ''


def test_client_log_exception(monkeypatch, caplog):
    org_url = "https://test.okta.com"
    token = "TOKEN"
    config = {'orgUrl': org_url, 'token': token,
              'logging': {'enabled': True, 'logLevel': logging.DEBUG}}
    client = OktaClient(config)

    class MockHTTPRequest():
        def __call__(self, **params):
            raise aiohttp.ClientConnectorCertificateError(
                ConnectionKey(host=org_url,
                              port=443,
                              is_ssl=True,
                              ssl=None,
                              proxy=None,
                              proxy_auth=None,
                              proxy_headers_hash=None),
                SSLCertVerificationError(1,
                                         '[SSL: CERTIFICATE_VERIFY_FAILED] certificate verify failed: '
                                         'unable to get local issuer certificate (_ssl.c:1123)'))

        async def __aenter__(self):
            return self

        async def __aexit__(self, exc_type, exc, tb):
            pass

        @staticmethod
        async def mock_response_text():
            return '[{"text": "mock response text"}]'

    mock_http_request = MockHTTPRequest()
    monkeypatch.setattr(aiohttp, 'request', mock_http_request)
    with caplog.at_level(logging.DEBUG):
        res, resp_body, error = asyncio.run(client.list_users())
        assert 'Cannot connect to host https://test.okta.com' in caplog.text



# flake8: noqa
"""
Copyright 2020 - Present Okta, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# AUTO-GENERATED! DO NOT EDIT FILE DIRECTLY
# SEE CONTRIBUTOR DOCUMENTATION

from okta.okta_object import OktaObject
from okta.okta_collection import OktaCollection
from okta.models import open_id_connect_application_type\
    as open_id_connect_application_type
from okta.models import open_id_connect_application_consent_method\
    as open_id_connect_application_consent_method
from okta.models import o_auth_grant_type\
    as o_auth_grant_type
from okta.models import open_id_connect_application_issuer_mode\
    as open_id_connect_application_issuer_mode
from okta.models import open_id_connect_application_settings_client_keys\
    as open_id_connect_application_settings_client_keys
from okta.models import open_id_connect_application_settings_refresh_token\
    as open_id_connect_application_settings_refresh_token
from okta.models import o_auth_response_type\
    as o_auth_response_type


class OpenIdConnectApplicationSettingsClient(
    OktaObject
):
    """
    A class for OpenIdConnectApplicationSettingsClient objects.
    """

    def __init__(self, config=None):
        super().__init__(config)
        if config:
            if "applicationType" in config:
                if isinstance(config["applicationType"],
                              open_id_connect_application_type.OpenIdConnectApplicationType):
                    self.application_type = config["applicationType"]
                elif config["applicationType"] is not None:
                    self.application_type = open_id_connect_application_type.OpenIdConnectApplicationType(
                        config["applicationType"].upper()
                    )
                else:
                    self.application_type = None
            else:
                self.application_type = None
            self.client_uri = config["clientUri"]\
                if "clientUri" in config else None
            if "consentMethod" in config:
                if isinstance(config["consentMethod"],
                              open_id_connect_application_consent_method.OpenIdConnectApplicationConsentMethod):
                    self.consent_method = config["consentMethod"]
                elif config["consentMethod"] is not None:
                    self.consent_method = open_id_connect_application_consent_method.OpenIdConnectApplicationConsentMethod(
                        config["consentMethod"].upper()
                    )
                else:
                    self.consent_method = None
            else:
                self.consent_method = None
            self.grant_types = OktaCollection.form_list(
                config["grantTypes"] if "grantTypes"\
                    in config else [],
                o_auth_grant_type.OAuthGrantType
            )
            self.initiate_login_uri = config["initiateLoginUri"]\
                if "initiateLoginUri" in config else None
            if "issuerMode" in config:
                if isinstance(config["issuerMode"],
                              open_id_connect_application_issuer_mode.OpenIdConnectApplicationIssuerMode):
                    self.issuer_mode = config["issuerMode"]
                elif config["issuerMode"] is not None:
                    self.issuer_mode = open_id_connect_application_issuer_mode.OpenIdConnectApplicationIssuerMode(
                        config["issuerMode"].upper()
                    )
                else:
                    self.issuer_mode = None
            else:
                self.issuer_mode = None
            if "jwks" in config:
                if isinstance(config["jwks"],
                              open_id_connect_application_settings_client_keys.OpenIdConnectApplicationSettingsClientKeys):
                    self.jwks = config["jwks"]
                elif config["jwks"] is not None:
                    self.jwks = open_id_connect_application_settings_client_keys.OpenIdConnectApplicationSettingsClientKeys(
                        config["jwks"]
                    )
                else:
                    self.jwks = None
            else:
                self.jwks = None
            self.logo_uri = config["logoUri"]\
                if "logoUri" in config else None
            self.policy_uri = config["policyUri"]\
                if "policyUri" in config else None
            self.post_logout_redirect_uris = OktaCollection.form_list(
                config["postLogoutRedirectUris"] if "postLogoutRedirectUris"\
                    in config else [],
                str
            )
            self.redirect_uris = OktaCollection.form_list(
                config["redirectUris"] if "redirectUris"\
                    in config else [],
                str
            )
            if "refreshToken" in config:
                if isinstance(config["refreshToken"],
                              open_id_connect_application_settings_refresh_token.OpenIdConnectApplicationSettingsRefreshToken):
                    self.refresh_token = config["refreshToken"]
                elif config["refreshToken"] is not None:
                    self.refresh_token = open_id_connect_application_settings_refresh_token.OpenIdConnectApplicationSettingsRefreshToken(
                        config["refreshToken"]
                    )
                else:
                    self.refresh_token = None
            else:
                self.refresh_token = None
            self.response_types = OktaCollection.form_list(
                config["responseTypes"] if "responseTypes"\
                    in config else [],
                o_auth_response_type.OAuthResponseType
            )
            self.tos_uri = config["tosUri"]\
                if "tosUri" in config else None
        else:
            self.application_type = None
            self.client_uri = None
            self.consent_method = None
            self.grant_types = []
            self.initiate_login_uri = None
            self.issuer_mode = None
            self.jwks = None
            self.logo_uri = None
            self.policy_uri = None
            self.post_logout_redirect_uris = []
            self.redirect_uris = []
            self.refresh_token = None
            self.response_types = []
            self.tos_uri = None

    def request_format(self):
        parent_req_format = super().request_format()
        current_obj_format = {
            "application_type": self.application_type,
            "client_uri": self.client_uri,
            "consent_method": self.consent_method,
            "grant_types": self.grant_types,
            "initiate_login_uri": self.initiate_login_uri,
            "issuer_mode": self.issuer_mode,
            "jwks": self.jwks,
            "logo_uri": self.logo_uri,
            "policy_uri": self.policy_uri,
            "post_logout_redirect_uris": self.post_logout_redirect_uris,
            "redirect_uris": self.redirect_uris,
            "refresh_token": self.refresh_token,
            "response_types": self.response_types,
            "tos_uri": self.tos_uri
        }
        parent_req_format.update(current_obj_format)
        return parent_req_format



from okta.cache.cache import Cache


class NoOpCache(Cache):
    """
    This is a disabled Cache Class where no operations occur
    in the cache.
    Implementing the okta.cache.cache.Cache abstract class.
    """

    def __init__(self):
        super()

    def get(self, key):
        """
        Nothing is returned.

        Arguments:
            key {str} -- Key to look for

        Returns:
            None -- No op cache doesn't contain any data to return
        """
        return None

    def contains(self, key):
        """
        False is returned

        Arguments:
            key {str} -- Key to look for

        Returns:
            False -- No data to return so key can never be in cache
        """
        return False

    def add(self, key, value):
        """
        This is a void method

        Arguments:
            key {str} -- Key in pair
            value {str} -- Val in pair
        """
        pass

    def delete(self, key):
        """This is a void method. No need to delete anything not contained.

        Arguments:
            key {str} -- Key to delete
        """
        pass

    def clear(self):
        """
        This is a void method. No need to clear when nothing's stored.
        """
        pass



import aiohttp
import asyncio
import json
import logging
import os
import xmltodict
from okta.errors.http_error import HTTPError
from okta.errors.okta_api_error import OktaAPIError
from okta.exceptions import HTTPException, OktaAPIException


logger = logging.getLogger('okta-sdk-python')


class HTTPClient:
    """
    This class is the basic HTTPClient for the Okta Client.
    Custom HTTP clients should inherit from this class.
    """
    raise_exception = False

    def __init__(self, http_config={}):
        # Get headers from Request Executor
        self._default_headers = http_config["headers"]
        # Create timeout for all HTTP requests
        self._timeout = aiohttp.ClientTimeout(
            total=http_config["requestTimeout"] if "requestTimeout" in
            http_config and http_config["requestTimeout"] > 0 else None
        )
        if "proxy" in http_config:
            self._proxy = self._setup_proxy(http_config["proxy"])
        else:
            self._proxy = None

    async def send_request(self, request):
        """
        This method fires HTTP requests

        Arguments:
            request {dict} -- This dictionary contains all information needed
            for the request.
            - HTTP method (as str)
            - Headers (as dict)
            - Request body (as dict)

        Returns:
            Tuple(aiohttp.RequestInfo, aiohttp.ClientResponse, JSON object)
            -- A tuple containing the request and response of the HTTP call
        """
        try:
            logger.debug(f"Request: {request}")
            # Set headers
            self._default_headers.update(request["headers"])
            # Prepare request parameters
            params = {'method': request['method'],
                      'url': request['url'],
                      'headers': self._default_headers}
            if request['data']:
                params['data'] = json.dumps(request['data'])
            json_data = request.get('json')
            # empty json param may cause issue, so include it if needed only
            # more details: https://github.com/okta/okta-sdk-python/issues/131
            if json_data:
                params['json'] = json_data
            if self._timeout:
                params['timeout'] = self._timeout
            if self._proxy:
                params['proxy'] = self._proxy
            # Fire request
            async with aiohttp.request(**params) as response:
                return (response.request_info,
                        response,
                        await response.text(),
                        None)
        except (aiohttp.ClientError, asyncio.TimeoutError) as error:
            # Return error if arises
            logger.exception(error)
            return (None, None, None, error)

    @staticmethod
    def check_response_for_error(url, response_details, response_body):
        """
        Checks HTTP response for errors in the response body

        Args:
            url (str): URL of response
            response_details (aiohttp.ClientResponse): Response object with
            details response_body (json): Response body in JSON format

        Returns:
            dict, (OktaAPIError or HTTPError): Tuple of: dict repr of the
            response (if no error), any error found
        """
        # Retrieve dictionary repr and response status code
        if response_details.content_type == "application/xml":
            formatted_response = xmltodict.parse(response_body)
        elif response_details.content_type == "application/json":
            formatted_response = json.loads(response_body)
        else:
            formatted_response = response_body

        status_code = response_details.status

        # check if call was succesful
        if 200 <= status_code <= 299:
            return (formatted_response, None)
        else:
            # create errors
            try:
                error = OktaAPIError(url, response_details, formatted_response)
                if HTTPClient.raise_exception:
                    raise OktaAPIException(formatted_response)
            except OktaAPIException:
                raise
            except Exception:
                error = HTTPError(url, response_details, formatted_response)
                if HTTPClient.raise_exception:
                    logger.exception(formatted_response)
                    raise HTTPException(formatted_response)
            logger.error(error)
            return (None, error)

    @staticmethod
    def format_binary_data(data):
        with aiohttp.MultipartWriter("mixed") as mpwriter:
            mpwriter.append(data)
        return mpwriter

    def _setup_proxy(self, proxy):
        proxy_string = ""

        if proxy is None:
            # check if env vars contain proxies
            if "HTTP_PROXY" in os.environ:
                proxy_string = os.environ["HTTP_PROXY"]
            if "HTTPS_PROXY" in os.environ:
                proxy_string = os.environ["HTTPS_PROXY"]
            return proxy_string if proxy_string != "" else None

        host = proxy["host"]
        port = int(proxy["port"]) if "port" in proxy else ""

        # config has credentials
        if "username" in proxy and "password" in proxy:
            username = proxy["username"]
            password = proxy["password"]
            proxy_string = f"http://{username}:{password}@{host}"
        else:
            proxy_string = f"http://{host}"

        if port:
            proxy_string += f":{port}/"

        return proxy_string if proxy_string != "" else None



"""
Copyright 2020 - Present Okta, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# AUTO-GENERATED! DO NOT EDIT FILE DIRECTLY
# SEE CONTRIBUTOR DOCUMENTATION

from okta.models.event_hook\
    import EventHook
from okta.utils import format_url
from okta.api_client import APIClient


class EventHookClient(APIClient):
    """
    A Client object for the EventHook resource.
    """

    def __init__(self):
        self._base_url = ""

    async def list_event_hooks(
            self
    ):
        """
        Args:
        Returns:
            list: Collection of EventHook instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/eventHooks
            """)

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, EventHook)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(EventHook(
                    self.form_response_body(item)
                    ))
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def create_event_hook(
            self, event_hook
    ):
        """
        Args:
            {event_hook}
        Returns:
            EventHook
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/eventHooks
            """)

        if isinstance(event_hook, dict):
            body = event_hook
        else:
            body = event_hook.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, EventHook)

        if error:
            return (None, response, error)

        try:
            result = EventHook(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def delete_event_hook(
            self, eventHookId
    ):
        """
        Args:
            event_hook_id {str}
        """
        http_method = "delete".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/eventHooks/{eventHookId}
            """)

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def get_event_hook(
            self, eventHookId
    ):
        """
        Args:
            event_hook_id {str}
        Returns:
            EventHook
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/eventHooks/{eventHookId}
            """)

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, EventHook)

        if error:
            return (None, response, error)

        try:
            result = EventHook(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def update_event_hook(
            self, eventHookId, event_hook
    ):
        """
        Args:
            event_hook_id {str}
            {event_hook}
        Returns:
            EventHook
        """
        http_method = "put".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/eventHooks/{eventHookId}
            """)

        if isinstance(event_hook, dict):
            body = event_hook
        else:
            body = event_hook.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, EventHook)

        if error:
            return (None, response, error)

        try:
            result = EventHook(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def activate_event_hook(
            self, eventHookId
    ):
        """
        Args:
            event_hook_id {str}
        Returns:
            EventHook
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/eventHooks/{eventHookId}/lifecycle/activate
            """)

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, EventHook)

        if error:
            return (None, response, error)

        try:
            result = EventHook(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def deactivate_event_hook(
            self, eventHookId
    ):
        """
        Args:
            event_hook_id {str}
        Returns:
            EventHook
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/eventHooks/{eventHookId}/lifecycle
                /deactivate
            """)

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, EventHook)

        if error:
            return (None, response, error)

        try:
            result = EventHook(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def verify_event_hook(
            self, eventHookId
    ):
        """
        Args:
            event_hook_id {str}
        Returns:
            EventHook
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/eventHooks/{eventHookId}/lifecycle/verify
            """)

        body = {}
        headers = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, EventHook)

        if error:
            return (None, response, error)

        try:
            result = EventHook(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)



from okta.errors.error import Error


class OktaAPIError(Error):
    def __init__(self, url, response_details, response_body):
        self.status = response_details.status
        self.error_code = response_body["errorCode"]
        self.error_summary = response_body.get("errorSummary", "")
        self.error_causes = response_body.get("errorCauses", [])
        error_causes_string = ". ".join(list(map(lambda x:
                                                 x.get("errorSummary", ""),
                                                 self.error_causes)))
        self.error_link = response_body["errorLink"]
        self.error_id = response_body["errorId"]

        self.url = url
        self.headers = response_details.headers
        self.stack = ""

        self.message = (f"Okta HTTP {self.status} {self.error_code} "
                        f"{self.error_summary}\n{error_causes_string}")



from okta.constants import FINDING_OKTA_APP_CRED, FINDING_OKTA_DOMAIN, \
    GET_OKTA_API_TOKEN

ERROR_MESSAGE_ORG_URL_MISSING = (
    "Your Okta URL is missing. You can copy "
    "your domain from the Okta Developer "
    "Console. Follow these instructions to"
    f" find it: {FINDING_OKTA_DOMAIN}"
)
ERROR_MESSAGE_ORG_URL_NOT_HTTPS = (
    "Your Okta URL must start with 'https'."
)
ERROR_MESSAGE_AUTH_MODE_INVALID = (
    "The AuthorizationMode configuration "
    "option must be one of: "
    "[SSWS, PrivateKey]. "
    "You provided the SDK with "
)
ERROR_MESSAGE_ORG_URL_YOUROKTADOMAIN = (
    "Replace {{yourOktaDomain}} with your Okta domain. "
    "You can copy your domain from the Okta Developer Console. "
    "Follow these instructions to find it: "
    f"{FINDING_OKTA_DOMAIN}"
)

ERROR_MESSAGE_ORG_URL_ADMIN = (
    "Your Okta domain should not contain -admin. "
)

ERROR_MESSAGE_ORG_URL_TYPO = (
    "It looks like there's a typo in your Okta domain."
)

ERROR_MESSAGE_API_TOKEN_MISSING = (
    "Your Okta API token is missing. "
    "You can generate one in the Okta Developer"
    " Console. Follow these instructions:"
    f" {GET_OKTA_API_TOKEN}"
)

ERROR_MESSAGE_API_TOKEN_DEFAULT = (
    "Replace {{apiToken}} with your Okta API token. "
    "You can generate one in the Okta Developer Console. "
    f"Follow these instructions: {GET_OKTA_API_TOKEN}"
)

ERROR_MESSAGE_CLIENT_ID_MISSING = (
    "Your client ID is missing. You can copy it from the "
    "Okta Developer Console in the details for the Application "
    "you created. Follow these instructions to find it: "
    f"{FINDING_OKTA_APP_CRED}"
)

ERROR_MESSAGE_CLIENT_ID_DEFAULT = (
    "Replace {{clientId}} with the client ID of your Application. "
    "You can copy it from the Okta Developer Console in the "
    "details for the Application you created. Follow these "
    f"instructions to find it: {FINDING_OKTA_APP_CRED}"
)

ERROR_MESSAGE_SCOPES_PK_MISSING = (
    "When using authorization mode 'PrivateKey', you must supply "
    "'okta.client.scopes' and 'okta.client.privateKey'"
)

ERROR_MESSAGE_429_MISSING_DATE_X_RESET = (
    "429 response must have the 'X-Rate-Limit-Reset' and 'Date' headers"
)

ERROR_MESSAGE_PROXY_MISSING_HOST = (
    "Please add a host URL to your proxy configuration."
)

ERROR_MESSAGE_PROXY_MISSING_AUTH = (
    "Proxy configuration must include BOTH your username and password "
    "if authentication is required"
)

ERROR_MESSAGE_PROXY_INVALID_PORT = (
    "Proxy port number must be a number, and between 1 and 65535 (inclusive)"
)
